"""AaC Plugin implementation module for the Version plugin."""

from os import path, makedirs, walk, remove
import importlib
from typing import Callable
from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
    OperationCancelled,
    ExecutionError,
    ExecutionMessage,
    MessageLevel,
)
from jinja2 import Environment, FileSystemLoader, Template
import black
from aac.context.language_context import LanguageContext
from aac.context.definition import Definition
from aac.in_out.parser._parse_source import parse
from aac.plugins.generate.helpers.python_helpers import (
    get_path_from_package,
    get_python_name,
)

plugin_name = "Generate"


def generate(  # noqa: C901
    aac_file: str,
    generator_file: str,
    code_output: str,
    test_output: str,
    doc_output: str,
    no_prompt: bool,
    force_overwrite: bool,
    evaluate: bool,
) -> ExecutionResult:
    """Generate content from your AaC architecture."""

    result = ExecutionResult(plugin_name, "generate", ExecutionStatus.SUCCESS, [])

    # setup directories
    code_out_dir = ""
    test_out_dir = ""
    doc_out_dir = ""
    try:
        code_out_dir, test_out_dir, doc_out_dir = get_output_directories(
            "AaC will generate code and tests in the following directories:",
            aac_file,
            code_output,
            test_output,
            doc_output,
            no_prompt,
        )
    except OperationCancelled as e:
        result.status_code = ExecutionStatus.OPERATION_CANCELLED
        result.add_message(ExecutionMessage(e.message, MessageLevel.INFO, None, None))
        return result

    # build out properties
    # the templates need data from the plugin model to generate code
    context = LanguageContext()
    parsed_definitions = parse(
        aac_file
    )  # we only want to parse, not load, to avoid chicken and egg issues

    # TO DO Consider surrounding this call to parse_and_load with a try except block to catch and handle the various exceptions that can be thrown
    generator_definitions = context.parse_and_load(generator_file)

    # go through each generator in the parsed_definitions
    for definition in generator_definitions:
        if definition.get_root_key() != "generator":
            continue
        generator = definition.instance

        # go through each generator source and get data from parsed_definitions based on from_source

        for source in generator.sources:
            source_data_definitions: list[Definition] = []
            for parsed_definition in parsed_definitions:
                if parsed_definition.get_root_key() == source.data_source:
                    source_data_definitions.append(parsed_definition)

            if not source_data_definitions or len(source_data_definitions) == 0:
                # no data for this particular generator
                continue

            # go through each generator template
            for template in source.templates:
                # figure out how to load func_dict into the jinja2 environment
                helper_functions: dict[str, Callable] = {}
                for helper in template.helper_functions:
                    helper_functions[helper.function] = get_callable(
                        helper.package, helper.module, helper.function
                    )

                # generate code using the template and source
                template_abs_path = path.abspath(
                    path.join(path.dirname(generator_file), template.template_file)
                )
                jinja_template = None
                if len(helper_functions) > 0:
                    jinja_template = load_template(template_abs_path, helper_functions)
                else:
                    jinja_template = load_template(template_abs_path)

                # loop over the parsed definitions and generate content for each
                for source_data_def in source_data_definitions:
                    source_data_structures = []
                    source_data_package = source_data_def.package
                    if not source.data_content:
                        # we'll just use the root structure
                        source_data_structures = [source_data_def.structure]
                    else:
                        # we've got to navigate the structure to get the right data
                        content_split = source.data_content.split(".")
                        if content_split[0] != source_data_def.get_root_key():
                            raise ExecutionError(
                                f"Invalid data_content for generator source {source.name}. The data_content must be the root key of the data source."
                            )
                        else:
                            source_data_structures = [source_data_def.structure]
                            for field_name in content_split:
                                new_source_data_structures = []
                                for structure in source_data_structures:
                                    if field_name not in structure:
                                        # it is possible that fields are optional and may not be present, so continue if not present
                                        continue
                                    field_value = structure[field_name]
                                    if isinstance(field_value, list):
                                        new_source_data_structures.extend(field_value)
                                    elif isinstance(field_value, dict):
                                        new_source_data_structures.append(field_value)
                                    else:
                                        raise ExecutionError(
                                            f"Invalid data_content {source.data_content} for generator source {source.name}. The data_content must be a field chain in the data source that represents a structure of data, not a primitive."
                                        )
                                source_data_structures = new_source_data_structures
                    for source_data_structure in source_data_structures:
                        jinja_output = jinja_template.render(source_data_structure)
                        output = jinja_output
                        if template.output_file_extension == "py":
                            output = black.format_str(jinja_output, mode=black.Mode())

                        # write output to files to the target in the template, respecting the overwrite indicator
                        root_out_dir = code_out_dir
                        if template.output_target == context.create_aac_enum(
                            "aac.lang.GeneratorOutputTarget", "TEST"
                        ):
                            root_out_dir = test_out_dir
                        elif template.output_target == context.create_aac_enum(
                            "aac.lang.GeneratorOutputTarget", "DOC"
                        ):
                            root_out_dir = doc_out_dir
                        file_name = source_data_def.name
                        if source.data_content:
                            name_extension = f"{source_data_structure['name'].replace(' ', '_').replace('-', '_')}"
                            file_name = f"{file_name}_{name_extension}"
                        output_file_path = get_output_file_path(
                            root_out_dir, template, source_data_package, file_name
                        )
                        # make sure the directory exists
                        output_dir = path.dirname(output_file_path)
                        if not path.exists(output_dir):
                            makedirs(output_dir)

                        if evaluate:
                            # write contents to an aac_evaluate file for user review
                            evaluate_file_path = f"{output_file_path}.aac_evaluate"
                            with open(evaluate_file_path, "w") as output_file:
                                output_file.write(output)
                                output_file.close()
                        else:
                            # write contents to output_file_path
                            if force_overwrite or template.overwrite in [
                                context.create_aac_enum(
                                    "aac.lang.OverwriteOption", "OVERWRITE"
                                )
                            ]:
                                if path.exists(output_file_path):
                                    backup_file(output_file_path)
                                with open(output_file_path, "w") as output_file:
                                    output_file.write(output)
                                    output_file.close()
                            elif template.overwrite in [
                                context.create_aac_enum(
                                    "aac.lang.OverwriteOption", "SKIP"
                                )
                            ]:
                                # this is for the skip option, so only write if file doesn't exist
                                if not path.exists(output_file_path):
                                    with open(output_file_path, "w") as output_file:
                                        output_file.write(output)
                                        output_file.close()
                                else:
                                    evaluate_file_path = (
                                        f"{output_file_path}.aac_evaluate"
                                    )
                                    with open(evaluate_file_path, "w") as output_file:
                                        output_file.write(output)
                                        output_file.close()

    return result


def clean(
    aac_file: str, code_output: str, test_output: str, doc_output: str, no_prompt: bool
) -> ExecutionResult:
    """Clean up generated code, tests, and docs."""
    # setup directories
    code_out_dir, test_out_dir, doc_out_dir = get_output_directories(
        "AaC will delete backup and eval files in the following directories:",
        aac_file,
        code_output,
        test_output,
        doc_output,
        no_prompt,
    )

    for dir in [code_out_dir, test_out_dir, doc_out_dir]:
        if path.exists(dir):
            for root, dirs, files in walk(dir):
                for file in files:
                    if file.endswith(".aac_backup") or file.endswith(".aac_evaluate"):
                        remove(path.join(root, file))

    return ExecutionResult(plugin_name, "generate", ExecutionStatus.SUCCESS, [])


def get_output_directories(
    message: str,
    aac_plugin_file: str,
    code_output: str,
    test_output: str,
    doc_output: str,
    no_prompt: bool,
) -> tuple[str, str, str]:
    """Returns the output directories for code, tests, and docs."""
    code_out: str = code_output
    test_out: str = test_output
    doc_out: str = doc_output

    if not code_output or len(code_output) == 0:
        aac_plugin_path = path.abspath(aac_plugin_file)
        if not path.exists(aac_plugin_file):
            raise ExecutionError(
                f"The provided AaC Plugin file does not exist: {aac_plugin_file}"
            )
        code_out = path.dirname(aac_plugin_path)

    if not test_output or len(test_output) == 0:
        test_out = path.dirname(path.abspath(aac_plugin_file))

    if not doc_output or len(doc_output) == 0:
        doc_out = path.dirname(path.abspath(aac_plugin_file))

    if not no_prompt:
        print(message)
        print(f"   code: {code_out}")
        print(f"   tests: {test_out}")
        print(f"   docs: {doc_out}")
        approval = "first"
        while approval not in [
            "y",
            "Y",
            "yes",
            "Yes",
            "YES",
            "n",
            "N",
            "no",
            "No",
            "NO",
        ]:
            approval = input("Do you want to continue? (y/n): ")
        if approval in ["y", "Y", "yes", "Yes", "YES"]:
            return (code_out, test_out, doc_out)
        else:
            raise OperationCancelled("User cancelled operation.")

    return (code_out, test_out, doc_out)


def get_output_file_path(
    root_output_directory: str,
    generator_template,
    source_package: str,
    source_name: str,
) -> str:
    """Returns the output file path for a generator template."""
    result = root_output_directory
    if generator_template.output_path_uses_data_source_package and source_package:
        context = LanguageContext()
        if generator_template.output_target == context.create_aac_enum(
            "aac.lang.GeneratorOutputTarget", "TEST"
        ):
            # this is a bit quirky, but turns out our test file path can't use the same structure as our source package or imports won't work properly
            # so for tests, we'll just add a tests_ prefix to the package name
            result = path.join(result, f"test_{get_path_from_package(source_package)}")
        else:
            result = path.join(result, get_path_from_package(source_package))
    file_name = ""
    if generator_template.output_file_prefix:
        file_name = generator_template.output_file_prefix
    if generator_template.output_file_name:
        file_name = f"{file_name}{generator_template.output_file_name}"
    elif source_name:
        file_name = f"{file_name}{get_python_name(source_name)}"
    if generator_template.output_file_suffix:
        file_name = f"{file_name}{generator_template.output_file_suffix}"
    file_name = f"{file_name}.{generator_template.output_file_extension}"
    result = path.join(result, file_name)
    return path.abspath(result)


def get_callable(package_name: str, file_name: str, function_name: str) -> Callable:
    """Returns a callable function from a package, file, and function name."""
    module = importlib.import_module(f"{package_name}.{file_name}")
    return getattr(module, function_name)


def load_template(
    template_abs_path: str, helper_functions: dict[str, Callable] = {}
) -> Template:
    """Load a jinja2 template from a file."""
    env = Environment(loader=FileSystemLoader("/"))
    env.globals.update(helper_functions)
    template = env.get_template(template_abs_path)
    return template


def backup_file(file_path: str) -> str:
    """Backs up a file by appending .aac_backup to the file name."""
    backup_file_path = f"{file_path}.aac_backup"
    with open(file_path, "r") as input_file:
        with open(backup_file_path, "w") as backup_file:
            backup_file.write(input_file.read())
    return backup_file_path



"""The LanguageContext is a singleton that holds the current state of the AaC language, including all definitions and plugin runners."""
from typing import Any, Type
from os.path import join, dirname
from aac.context.language_error import LanguageError
from aac.execute.plugin_manager import get_plugin_manager
from aac.execute.plugin_runner import PluginRunner
from aac.context.definition import Definition
from aac.in_out.parser._parse_source import parse
from aac.context.definition_parser import DefinitionParser

AAC_LANG_FILE_NAME = "../aac.aac"
AAC_LANG_FILE_PATH = join(dirname(__file__), AAC_LANG_FILE_NAME)


class LanguageContext(object):
    """A singleton class that holds the current state of the AaC language."""

    def __new__(cls):
        """Create a new instance of the LanguageContext singleton class."""
        if not hasattr(cls, "context_instance"):
            cls.context_instance = super(LanguageContext, cls).__new__(cls)
            cls.context_instance.definitions = set()
            cls.context_instance.fully_qualified_name_to_definition: dict[
                str, Definition
            ] = {}
            cls.context_instance.fully_qualified_name_to_class: dict[str, Any] = {}
            cls.context_instance.plugin_runners = {}

            # load and initialize the AaC language
            cls.context_instance.parse_and_load(AAC_LANG_FILE_PATH)

            # load plugins
            get_plugin_manager().hook.register_plugin()

        return cls.context_instance

    def get_aac_core_file_path(self) -> str:
        """
        Function to return the AaC language file path.

        Returns:
            The filepath to the core AaC definitions file as a string.
        """
        return AAC_LANG_FILE_PATH

    def get_aac_core_as_yaml(self) -> str:
        """
        Function to return the AaC language as a YAML string.

        Returns:
            The core AaC definitions file as a string in YAML format.
        """
        with open(AAC_LANG_FILE_PATH) as aac_file:
            return aac_file.read()

    def get_aac_core_definitions(self) -> list[Definition]:
        """
        Function to return the definitions for the AaC language.

        Returns:
            A list of definitions in the core AaC definitions file.
        """
        return self.parse_and_load(AAC_LANG_FILE_PATH)

    def _get_aac_generated_class(self, name: str) -> Type:
        definitions = self.get_definitions_by_name(name)
        if len(definitions) != 1:
            raise LanguageError(
                f"_get_aac_generated_class unable to identify unique definition for name '{name}'.  Found: {[definition.name for definition in definitions]}",
                definitions[0].source.uri
            )

        aac_class = self.context_instance.fully_qualified_name_to_class[
            definitions[0].get_fully_qualified_name()
        ]
        if not aac_class:
            raise LanguageError(
                f"_get_aac_generated_class unable to identify generated class for name '{name}' with fully_qualified_name '{definitions[0].get_fully_qualified_name()}'",
                definitions[0].source.uri
            )

        return aac_class

    def is_aac_instance(self, obj: Any, name: str) -> bool:
        """
        Function to determine if an object is an instance of an AaC class.

        Args:
            obj (Any): Object to check.
            name (str):  The AaC class name to be used for comparison.

        Returns:
            A boolean which equals true if the object is an instance of the specified class.
        """
        return isinstance(obj, self._get_aac_generated_class(name))

    def create_aac_object(self, aac_type_name: str, attributes: dict) -> Any:
        """
        Function to create a Python object representation of an AaC class and its attributes.

        Args:
            aac_type_name (str): The name of the AaC class.
            attributes (dict): A dictionary of attributes for the AaC class.

        Returns:
            A Python object representing the AaC class.

        Raises:
            LanguageError: When a unique definition for the AaC type is not found, an error message detailing the issue is generated.
        """
        definitions = self.get_definitions_by_name(aac_type_name)
        if len(definitions) != 1:
            raise LanguageError(
                f"Unable to identify unique definition for '{aac_type_name}'.  Found {[definition.name for definition in definitions]}",
                definitions[0].source.uri
            )
        aac_class = self._get_aac_generated_class(aac_type_name)
        result = aac_class()
        field_names = [field.name for field in definitions[0].instance.fields]
        for attribute_name, attribute_value in attributes.items():
            if attribute_name not in field_names:
                raise LanguageError(
                    f"Found undefined field name '{attribute_name}'",
                    definitions[0].source.uri
                )
            setattr(result, attribute_name, attribute_value)
        return result

    def create_aac_enum(self, aac_enum_name: str, value: str) -> Any:
        """
        Function to create a Python enum for an AaC enum class and value.

        Args:
            aac_enum_name (str): The name of the AaC Enum class.
            value (str): The value for the enum.

        Returns:
            A Python enum set to the specified value.

        Raises:
            LanguageError: When a unique definition for the AaC Enum is not found, an error message detailing the issue is generated.
        """
        definitions = self.get_definitions_by_name(aac_enum_name)
        if len(definitions) > 1:
            raise LanguageError(
                f"Unable to identify unique definition for '{aac_enum_name}'.  Found {[definition.name for definition in definitions]}",
                definitions[0].source.uri
            )
        if len(definitions) < 1:
            raise LanguageError(
                f"Unable to identify definition for '{aac_enum_name}'",
                "No file to reference"
            )
        aac_class = self._get_aac_generated_class(aac_enum_name)
        try:
            return getattr(aac_class, value)
        except ValueError:
            raise LanguageError(
                f"{value} is not a valid value for enum {aac_enum_name}",
                definitions[0].source.uri
            )

    def parse_and_load(self, arg: str) -> list[Definition]:
        """
        Convenience function that parses a file or string and loads the definitions into the context.

        Args:
            arg (str): An AaC definition file.  Can be a string in YAML format or a filepath.

        Returns:
            A list of definition objects which have been loaded into the Language Context.
        """
        parsed_definitions = parse(arg)
        parser = DefinitionParser()

        return parser.load_definitions(self, parsed_definitions)

    def remove_definitions(self, definitions: list[Definition]) -> None:
        """
        Remove the given definitions from the context.

        Args:
            definitions (list[Definition]): A list of Definitions to be removed from the Language Context.
        """
        for definition in definitions:
            definition.source.is_loaded_in_context = False
            self.context_instance.definitions.remove(definition)
            del self.context_instance.fully_qualified_name_to_definition[
                f"{definition.package}.{definition.name}"
            ]

    def get_definitions(self) -> list[Definition]:
        """
        Get all the definitions.

        Returns:
            A list of all Definitions currently loaded into the Language Context.
        """
        return list(self.context_instance.fully_qualified_name_to_definition.values())

    def get_definitions_by_name(self, name: str) -> list[Definition]:
        """
        Get all the definitions with a given name.

        Args:
            name (str): The name of the definition(s).

        Returns:
            A list of definitions with the given name.
        """
        result = []
        search_name = name
        if "." not in name:
            search_name = f".{name}"
        for definition in self.get_definitions():
            if definition.get_fully_qualified_name().endswith(search_name.replace(" ", "")):
                result.append(definition)
        return result

    def get_definitions_by_root(self, root_key: str) -> list[Definition]:
        """
        Get all the definitions with a given root key.

        Args:
            root_key (str): A root key to search for.

        Returns:
           A list of definitions with the given root key.
        """
        result = []
        for definition in self.get_definitions():
            if definition.get_root_key() == root_key:
                result.append(definition)
        return result

    def get_defining_schema_for_root(self, root_key: str) -> Definition:
        """
        Get the defining schema for a given root key.

        Args:
            root_key (str): The root key for the schema definition.

        Returns:
            The definition of the root schema.

        Raises:
            LanguageError: When no definition is found for the defining schema, an error message detailing the issue is generated.
        """
        current_definition: Definition = None
        for definition in self.get_definitions():
            current_definition = definition
            if definition.get_root_key() == "schema":
                if definition.instance.root == root_key:
                    return definition
        raise LanguageError(
            message=f"Could not find defining schema for root key: {root_key}",
            location=current_definition.source.uri
        )

    def register_plugin_runner(self, runner: PluginRunner) -> None:
        """
        Register a plugin runner.

        Args:
            runner (PluginRunner): A plugin runner to be registered.
        """
        if runner.get_plugin_name() not in self.context_instance.plugin_runners:
            self.context_instance.plugin_runners[runner.get_plugin_name()] = runner
        else:
            print(f"Plugin {runner.get_plugin_name()} already registered.")

    def get_plugin_runners(self) -> list[PluginRunner]:
        """
        Get all the plugin runners.

        Returns:
            A list of all registered plugin runners.
        """
        return list(self.context_instance.plugin_runners.values())

    def get_primitives(self) -> list[Definition]:
        """
        Get all the primitive definitions.

        Returns:
            A list of all primitive definitions.
        """
        return self.get_definitions_by_root("primitive")

    def get_python_type_from_primitive(self, primitive_name: str) -> str:
        """
        Get the Python type from a primitive name.

        Args:
            primitive_name (str): name of the primitive type.

        Returns:
            The Python type equivalent to the declared primitive type.

        Raises:
            LanguageError: When no primitive type is found with the given name, an error message detailing the issue is generated.
        """
        primitives = self.get_definitions_by_name(primitive_name)
        if len(primitives) > 1:
            raise LanguageError(
                message=f"Could not find unique primitive type: {primitive_name} - discovered {[primitive.name for primitive in primitives]}",
                location=primitives[0].source.uri
            )
        if len(primitives) < 1:
            raise LanguageError(
                message=f"Could not find primitive type: {primitive_name}",
                location="No file to reference"
            )
        return primitives[0].instance.python_type

    def is_extension_of(self, check_me: Definition, package: str, name: str) -> bool:
        """
        Check to see if a given definition extends from a given package and name.

        Args:
            check_me (Definition): The definition to be checked if it is an extension.
            package (str): The package name of the definitions to be checked against.
            name (str): The name of the definitions to be checked against.

        Returns:
            A boolean value which equals True if the given definition is an extension, and equals False if it is not.
        """
        definitions_of_type = self.get_definitions_of_type(package, name)
        return check_me in definitions_of_type

    def _recurse_type_inheritance(self, check_me: Definition) -> list[Definition]:
        result = []
        if "extends" in list(vars(check_me.instance).keys()):
            for extension in check_me.instance.extends:
                ext_package = extension.package
                ext_name = extension.name
                parent_definition = (
                    self.context_instance.fully_qualified_name_to_definition[
                        f"{ext_package}.{ext_name}"
                    ]
                )
                result.append(parent_definition)
                result.extend(self._recurse_type_inheritance(parent_definition))
        return result

    def get_definitions_of_type(self, package: str, name: str) -> list[Definition]:
        """
        Search the language context to find definitions that match a given package and name.

        Args:
            package (str): The package name of the type.
            name (str): The name of the type.

        Returns:
            A list of definitions of the specified type within the specified package.

        Raises:
            LanguageError: When no definition is found with the given name within the given package, an error message detailing the issue is generated.
        """
        result = []
        # first make sure we can find the definition that defines the type
        if (
            f"{package}.{name}"
            not in self.context_instance.fully_qualified_name_to_definition
        ):
            raise LanguageError(f"Could not find definition for {package}.{name}", "No file to reference")
        # now get all the definitions that extend from the definition
        definition = self.context_instance.fully_qualified_name_to_definition[
            f"{package}.{name}"
        ]
        result.append(definition)
        # loop through all definitions and see if the type we care about is in it's inheritance tree
        for item in self.get_definitions():
            inheritance_types = self._recurse_type_inheritance(item)
            if f"{package}.{name}" in [
                f"{definition.package}.{definition.name}"
                for definition in inheritance_types
            ]:
                result.append(item)
        return result

    def _traverse_field_chain(
        self, structure_list: list, field_chain: list[str]
    ) -> list:
        current_structures = structure_list
        search_depth = len(field_chain)
        current_depth = 0
        values_found = []
        for field_name in field_chain:
            next_structures = []
            for structure in current_structures:
                if field_name in structure:
                    if isinstance(structure[field_name], list):
                        if current_depth == search_depth - 1:
                            values_found.extend(structure[field_name])
                        else:
                            next_structures.extend(structure[field_name])
                    else:
                        if (
                            current_depth == search_depth - 1
                        ):  # This just means we're at the end of the field chain and should pull the value
                            values_found.append(structure[field_name])
                        else:
                            next_structures.append(structure.get(field_name, []))
            current_structures = next_structures
            current_depth += 1

        return values_found

    def get_values_by_field_chain(self, search_term: str) -> list:
        """
        Find values from the language context using a dot notation field chain.

        Args:
            search_term (str): A dot notation field chain. (i.e. root.definition.field)

        Returns:
            A list of values from the specified fields.
        """
        result: list = []
        root_key = search_term.split(".")[0]
        candidate_values = self.get_definitions_by_root(root_key)
        result = self._traverse_field_chain(
            [definition.structure for definition in candidate_values],
            search_term.split("."),
        )
        return result



from unittest import TestCase
from tempfile import TemporaryDirectory

from aac.in_out.files.find import find_aac_files, is_aac_file
from aac.context.language_context import LanguageContext

YAML_DOCUMENT_EXTENSION = "yaml"
AAC_DOCUMENT_EXTENSION = "aac"

class TestFindFiles(TestCase):

    def create_test_file(self, path: str, content: str) -> str:
        with open(path, "w") as test_file:
            test_file.write(content)
        return path
    
    def test_is_aac_file_with_valid_yaml_file(self):
        context = LanguageContext()
        test_file_content = context.get_aac_core_as_yaml()
        with TemporaryDirectory() as temp_dir:
            test_yaml = self.create_test_file(f"{temp_dir}/test.{YAML_DOCUMENT_EXTENSION}", test_file_content)
            actual_result = is_aac_file(test_yaml)

            self.assertTrue(actual_result)

    def test_is_aac_file_with_valid_aac_file(self):
        context = LanguageContext()
        test_file_content = context.get_aac_core_as_yaml()
        with TemporaryDirectory() as temp_dir:
            test_yaml = self.create_test_file(f"{temp_dir}/test.{AAC_DOCUMENT_EXTENSION}", test_file_content)
            actual_result = is_aac_file(test_yaml)

            self.assertTrue(actual_result)

    def test_is_aac_file_with_invalid_yaml_file(self):
        with TemporaryDirectory() as temp_dir:
            test_yaml = self.create_test_file(f"{temp_dir}/test.{YAML_DOCUMENT_EXTENSION}", VALID_NON_AAC_YAML_CONTENT)
            actual_result = is_aac_file(test_yaml)

            self.assertFalse(actual_result)

    def test_is_aac_file_with_invalid_aac_file(self):
        with TemporaryDirectory() as temp_dir:
            test_yaml = self.create_test_file(f"{temp_dir}/test.{AAC_DOCUMENT_EXTENSION}", VALID_NON_AAC_YAML_CONTENT)
            actual_result = is_aac_file(test_yaml)

            self.assertFalse(actual_result)

    def test_is_aac_file_with_invalid_file_extension(self):
        context = LanguageContext()
        test_file_content = context.get_aac_core_as_yaml()
        with TemporaryDirectory() as temp_dir:
            test_yaml = self.create_test_file(f"{temp_dir}/test.notYamlOrAac", test_file_content)
            actual_result = is_aac_file(test_yaml)

            self.assertFalse(actual_result)

    def test_find_aac_files(self):
        context = LanguageContext()
        with TemporaryDirectory() as l1_temp_dir:
            l1_aac_file_aac = self.create_test_file(f"{l1_temp_dir}/l1_aac_file.aac", context.get_aac_core_as_yaml())
            l1_invalid_file_aac = self.create_test_file(f"{l1_temp_dir}/l1_invalid_file.aac", VALID_NON_AAC_YAML_CONTENT)
        
            with TemporaryDirectory(dir=l1_temp_dir) as l2_temp_dir:
                l2_aac_file_yaml = self.create_test_file(f"{l2_temp_dir}/l2_aac_file.yaml", context.get_aac_core_as_yaml())
                l2_invalid_file_yaml = self.create_test_file(f"{l2_temp_dir}/l2_invalid_file.yaml", VALID_NON_AAC_YAML_CONTENT)

                with TemporaryDirectory(dir=l2_temp_dir) as l3_temp_dir:
                    l3_aac_file_aac = self.create_test_file(f"{l3_temp_dir}/l3_aac_file.yaml", context.get_aac_core_as_yaml())
                    l3_invalid_file_aac = self.create_test_file(f"{l3_temp_dir}/l3_invalid_file.yaml", VALID_NON_AAC_YAML_CONTENT)

                    expected_result = [l1_aac_file_aac, l2_aac_file_yaml, l3_aac_file_aac]
                    actual_result = find_aac_files(l1_temp_dir)

                    actual_result_file_paths = [file.uri for file in actual_result]
                    self.assertListEqual(expected_result, actual_result_file_paths)
                    self.assertNotIn(l1_invalid_file_aac, actual_result_file_paths)
                    self.assertNotIn(l2_invalid_file_yaml, actual_result_file_paths)
                    self.assertNotIn(l3_invalid_file_aac, actual_result_file_paths)


NON_YAML_FILE_CONTENT = """
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut
labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco
laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in
voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat
non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
"""

VALID_NON_AAC_YAML_CONTENT = """
apiVersion: xl-deploy/v1
kind: Infrastructure
spec:
- name: Infrastructure/Apache  host
  type: overthere.SshHost
  os: UNIX
  address: tomcat-host.local
  username: tomcatuser
- name: Infrastructure/local-docker
  type: docker.Engine
  dockerHost: http://dockerproxy:2375
- name: aws
  type: aws.Cloud
  accesskey: YOUR ACCESS KEY
  accessSecret: YOUR SECRET
"""



"""Definition Parser class for AaC, contains a load_definition function that handles loading of definition files."""
from typing import Any, Type
from enum import Enum, auto
from aac.context.language_error import LanguageError
from aac.context.definition import Definition
from aac.context.lexeme import Lexeme
from aac.context.util import get_python_module_name, get_python_class_name


class DefinitionParser():
    """Definition Parser class, responsible for loading definition files."""

    def find_definitions_by_name(self, name: str) -> list[Definition]:
        """
        Method to find a definition by name.

        Args:
            name (str): The name of the definition being searched for.

        Returns:
            The definition with the given name.
        """
        result = []
        for definition in self.context.get_definitions():
            if definition.name == name:
                result.append(definition)
        # if we didn't find any definitions in the context, check the parsed definitions
        if len(result) == 0:
            for definition in self.parsed_definitions:
                if definition.name == name:
                    result.append(definition)
        return result

    def get_location_str(self, lexeme_value: str, lexemes: list[Lexeme]) -> str:
        """
        Method to find the file name and line number for a requested Lexeme value.

        Args:
            lexeme_value (str): The Lexeme to match.
            lexemes (list[Lexeme]): A list of definition Lexemes.

        Returns:
            The file name and line number of the requested Lexeme value.
        """
        lexeme = []
        for lex in lexemes:
            # This is just a way to account for different variations of true and false.
            if lexeme_value is True and (lex.value == "True" or lex.value == "true" or lex.value == "yes"):
                lexeme.append(lex)
            if lexeme_value is False and (lex.value == "False" or lex.value == "false"):
                lexeme.append(lex)
            # The normal found case.  lexeme_value is being cast as a string because all values are stored in lexemes as strings.
            elif lex.value == str(lexeme_value):
                lexeme.append(lex)
        location_str = (
            "Unable to identify source and location"  # this is the 'not found' case
        )
        if len(lexeme) == 1:  # this is the single match case
            source_str = lexeme[0].source
            line_str = lexeme[0].location.line + 1
            location_str = f"File: {source_str}  Line: {line_str}"
        elif len(lexeme) > 1:  # this is the ambiguous match case
            # check to see if location source is the same for all matches
            if all(
                [
                    lexeme[0].source == lexeme[i].source
                    for i in range(1, len(lexeme))
                ]
            ):
                source_str = lexeme[0].source
                location_str = f"File: {source_str}  Possible Lines: {', '.join([str(lex.location.line+1) for lex in lexeme])}"
            else:  # if not, just list each possible location
                location_str = "Unable to identify unique location - "
                for lex in lexeme:
                    location_str += (
                        f"File: {lex.source}  Line: {lex.location.line+1}  "
                    )
        return location_str

    def get_inheritance_parents(self, definition: Definition) -> list[Type]:
        """
        Looks up the inheritance parent classes for the given definition and returns them as a list of Python classes.

        Args:
            definition (Definition): The definition whose parent class(es) is being searched for.

        Returns:
            The parent class(es) as a list of Python classes.
        """
        inheritance_parents = []

        if "extends" in definition.structure[definition.get_root_key()]:
            for parent in definition.structure[definition.get_root_key()][
                "extends"
            ]:
                # I need to find the definition referenced by the extends
                parent_package = parent["package"]
                parent_name = parent["name"]
                parent_fully_qualified_name = ""
                try:
                    parent_fully_qualified_name = f"{get_python_module_name(parent_package)}.{get_python_class_name(parent_name)}"
                except LanguageError as e:
                    raise LanguageError(
                        f"Failed to establish parent fully qualified name from parent_package {parent_package} and parent_name {parent_name}: {e.message}", self.get_location_str(parent_name, definition.lexemes)
                    )

                if (
                    parent_fully_qualified_name
                    in self.context.context_instance.fully_qualified_name_to_class
                ):
                    inheritance_parents.append(
                        self.context.context_instance.fully_qualified_name_to_class[
                            parent_fully_qualified_name
                        ]
                    )
                else:
                    # there is a chance that processing order just means we haven't gotten to the parent yet
                    parent_definition = None
                    if (
                        parent_fully_qualified_name
                        in self.fully_qualified_name_to_definition
                    ):
                        parent_definition = self.fully_qualified_name_to_definition[
                            parent_fully_qualified_name
                        ]

                    if not parent_definition:
                        raise LanguageError(
                            f"Cannot find parent definition {parent_fully_qualified_name} for {definition.name}",
                            self.get_location_str(parent_name, definition.lexemes),
                        )

                    if parent_definition.get_root_key() == "schema":
                        inheritance_parents.append(
                            self.create_schema_class(parent_definition)
                        )
                    else:
                        # AaC only supports schema inheritance
                        raise LanguageError(
                            f"AaC extension is only supported for schema.  Unable to create parent class with AaC root: {parent_definition.get_root_key()}",
                            self.get_location_str(parent_name, definition.lexemes),
                        )

        return inheritance_parents

    def create_enum_class(self, enum_definition: Definition) -> Enum:
        """
        Creates an enum class from a given enum definition.

        Args:
            enum_definition (Definition): An enum definition to convert to a class.

        Returns:
            The created class.
        """
        if enum_definition.get_root_key() != "enum":
            raise LanguageError(
                f"Definition {enum_definition.name} is not an enum",
                self.get_location_str(
                    enum_definition.get_root_key(), enum_definition.lexemes
                )
            )

        try:
            fully_qualified_name = enum_definition.get_fully_qualified_name()
        except LanguageError as e:
            raise LanguageError(
                f"Failed to create fully qualified name for definition {enum_definition.name}: {e.message}",
                self.get_location_str(enum_definition.name, enum_definition.lexemes),
            )

        if (
            fully_qualified_name
            in self.context.context_instance.fully_qualified_name_to_class
        ):
            # we've already created the class, so nothing to do here
            return self.context.context_instance.fully_qualified_name_to_class[
                fully_qualified_name
            ]

        # Note:  trying to allow extension with Enums fails, so I just removed it
        #        but we can revisit and try to find a solution in the future if needed

        values = {}
        if "values" in enum_definition.structure["enum"]:
            for value in enum_definition.structure["enum"]["values"]:
                values[value] = auto()
        # create the enum class
        instance_class = None
        try:
            instance_class = Enum(
                enum_definition.get_python_class_name(),
                values,
                module=enum_definition.get_python_module_name(),
            )
        except LanguageError as e:
            raise LanguageError(
                f"Failed to create Enum instance_class for {enum_definition.name}: {e.message}",
                self.get_location_str(enum_definition.name, enum_definition.lexemes),
            )
        self.context.context_instance.fully_qualified_name_to_class[
            fully_qualified_name
        ] = instance_class
        return instance_class

    def add_field_to_class(self, field: dict, instance_class: Type, schema_definition: Definition) -> Type:
        """
        Creates an attribute within the instance class and adds the field name and type to it.

        Args:
            field (dict): The field being loaded into the instance class.
            instance_class (Type): The instance class being loaded with field definitions.
            schema_definition (Definition): The schema definition containing the fields being iterated through and loaded.

        Returns:
            the instance class after being loaded with field definitions.
        """
        field_name = field["name"]
        field_type = field["type"]
        is_list = False

        clean_field_type = field_type
        if field_type.endswith("[]"):
            is_list = True
            clean_field_type = field_type[:-2]
        if "(" in clean_field_type:
            clean_field_type = clean_field_type[: clean_field_type.find("(")]

        # let's make sure the type of the field is known, or create it if it's not
        potential_definitions = self.find_definitions_by_name(clean_field_type)
        if len(potential_definitions) != 1:
            if len(potential_definitions) == 0:
                raise LanguageError(
                    f"Could not find AaC definition for type {clean_field_type} while loading {schema_definition.name}",
                    self.get_location_str(field_type, schema_definition.lexemes),
                )
            else:
                raise LanguageError(
                    f"Discovered multiple AaC definitions for type {clean_field_type} while loading {schema_definition.name}.  You may need to add a package name to differentiate.",
                    self.get_location_str(field_type, schema_definition.lexemes),
                )

        parsed_definition = potential_definitions[0]

        self.create_schema_class(parsed_definition)

        # since python is dynamically typed, we really don't have to worry about setting a type when we create the field
        # we just need to make sure a reasonable default value is used, so for us that means an empty list or None
        # Question:  is there ever a case where we may need a dict value?
        if is_list:
            setattr(instance_class, field_name, [])
        else:
            setattr(instance_class, field_name, None)
        return instance_class

    def create_instance_class(self, inheritance_parents: list, schema_definition: Definition) -> Type:
        """
        Method to create an instance class type for the given definition.

        Args:
            inheritance_parents (list[Type]): Types that this definition inherits from.
            schema_definition (Definition): The definition being converted to an instance class.

        Returns:
            The created instance class.
        """
        if len(inheritance_parents) == 0:
            base_content = [object]
        else:
            base_content = inheritance_parents

        try:
            instance_class = type(
                schema_definition.get_python_class_name(),
                tuple(base_content),
                {"__module__": schema_definition.get_python_module_name()}
            )
        except LanguageError as e:
            raise LanguageError(
                f"Failed to create instance_class for {schema_definition.name}: {e.message}",
                self.get_location_str(
                    schema_definition.name, schema_definition.lexemes
                )
            )
        return instance_class

    def create_schema_class(self, schema_definition: Definition) -> Type:
        """
        Creates a schema class from a given schema definition.

        Args:
            schema_definition (Definition): A schema definition to convert to a class.

        Returns:
            The created class.
        """
        try:
            fully_qualified_name = schema_definition.get_fully_qualified_name()
        except LanguageError as e:
            raise LanguageError(
                f"Failed to create fully qualified name for definition {schema_definition.name}: {e.message}",
                self.get_location_str(schema_definition.name, schema_definition.lexemes),
            )

        if schema_definition.get_root_key() == "primitive":
            # this is a primitive, so there's no structure to create, just return the python type
            return eval(self.primitive_name_to_py_type[schema_definition.name])
        elif schema_definition.get_root_key() == "enum":
            # this is an enum, so create the enum class
            return self.create_enum_class(schema_definition)
        elif schema_definition.get_root_key() != "schema":
            raise LanguageError(
                f"Definition {schema_definition.name} is not a schema",
                self.get_location_str(
                    schema_definition.get_root_key(), schema_definition.lexemes
                )
            )

        if (
            fully_qualified_name
            in self.context.context_instance.fully_qualified_name_to_class
        ):
            # we've already created the class, so nothing to do here
            return self.context.context_instance.fully_qualified_name_to_class[
                fully_qualified_name
            ]

        inheritance_parents = self.get_inheritance_parents(schema_definition)
        instance_class = self.create_instance_class(inheritance_parents, schema_definition)

        # now add the fields to the class
        for field in schema_definition.structure["schema"]["fields"]:
            instance_class = self.add_field_to_class(field, instance_class, schema_definition)

        # finally store the class in the context
        self.context.context_instance.fully_qualified_name_to_class[
            fully_qualified_name
        ] = instance_class
        return instance_class

    def create_object_instance(self, type_class: Type, fields: dict) -> Any:
        """
        Creates an instance object from the given fields.

        Args:
            type_class (Type): The class created from the field type.
            fields (dict): Given fields to create instances from.

        Returns:
            The created instance object.
        """
        result = type_class()
        for field_name, field_value in fields.items():
            setattr(result, field_name, field_value)
        return result

    def get_defined_fields(self, package: str, name: str) -> list[str]:
        """
        Returns a list of defined fields for the given definition.

        Args:
            package (str): The definition package.
            name (str): The definition name.

        Returns:
            A list of definition fields.
        """
        result = []
        defining_definition = None
        for definition in self.context.get_definitions() + self.parsed_definitions:
            if definition.name == name and definition.package == package:
                defining_definition = definition
                break
        if (
            "extends"
            in defining_definition.structure[defining_definition.get_root_key()]
        ):
            for parent in defining_definition.structure[
                defining_definition.get_root_key()
            ]["extends"]:
                result.extend(self.get_defined_fields(parent["package"], parent["name"]))
        if "fields" in defining_definition.structure[definition.get_root_key()]:
            result.extend(
                [
                    field["name"]
                    for field in definition.structure[definition.get_root_key()][
                        "fields"
                    ]
                ]
            )
        return result

    def populate_sub_fields(self, subfields: dict, defining_definition: Definition, item: dict, lexemes: list[Lexeme], definition: Definition) -> dict:
        """
        Method used to populate sub-fields in an instance class field.

        Args:
            subfields (dict): dictionary containing relevant sub-fields.
            defining_definition (Definition): The definition of the field type.
            item (dict): The current field definition which contains sub-fields.
            lexemes (List[Lexeme]): A list of definition Lexemes.
            definition (Definition): The overall definition containing the field.

        Returns:
            The populated dictionary of sub-fields.
        """
        for subfield in defining_definition.structure["schema"][
            "fields"
        ]:
            subfield_name = subfield["name"]
            subfield_type = subfield["type"]
            subfield_is_required = False
            if "is_required" in subfield:
                subfield_is_required = subfield["is_required"]
            subfield_value = None
            if subfield_name in item:
                subfield_value = item[subfield_name]
            else:
                if "default" in subfield:
                    # let's see if we need to cast the value
                    subfield_default_str = subfield["default"]
                    if isinstance(subfield_default_str, str):
                        type_map = {
                            "int": int,
                            "number": float,
                            "bool": lambda x: x.lower()
                            in ("yes", "true", "t", "1"),
                            "string": str,
                        }
                        if subfield_type in type_map:
                            subfield_value = type_map[
                                subfield_type
                            ](subfield_default_str)
                    else:
                        subfield_value = subfield_default_str
            # we need to eliminate previously covered lexemes, so go through the list until we find subfield_name then add it and everything else
            found = False
            sub_lexemes = []
            for lex in lexemes:
                if lex.value == subfield_name:
                    found = True
                    sub_lexemes.append(lex)
                if found:
                    sub_lexemes.append(lex)
            subfields[subfield_name] = self.create_field_instance(
                subfield_name,
                subfield_type,
                subfield_is_required,
                subfield_value,
                sub_lexemes,
                definition,
            )
        return subfields

    def field_instance_creator_list(self, lexemes: list, field_value: Any, field_name: str, defining_definition: Definition, instance: list, instance_class: Type, definition: Definition) -> list:
        """
        Method to create instance fields from list type fields.

        Args:
            lexemes (List[Lexeme]): A list of definition Lexemes.
            field_value (Any): Value stored in the field.
            field_name (str): The name of the field.
            defining_definition (Definition): Definition of the field type.
            instance (list[Any]: The instance that will contain the created field)
            instance_class (Type): The instance type class
            definition (Definition): Definition containing the field being checked.

        Returns:
            The updated instance containing the created field.
        """
        for item in field_value:
            if not isinstance(item, dict):
                raise LanguageError(
                    f"Invalid parsed value for field '{field_name}'.  Expected type 'dict', but found '{type(item)}'. Value = {item}",
                    self.get_location_str(field_name, lexemes),
                )
            # go through the fields and create instances for each
            subfields = {}
            if "fields" not in defining_definition.structure["schema"]:
                raise LanguageError(
                    f"Schema '{defining_definition.name}' does not contain any fields.",
                    self.get_location_str(field_name, lexemes),
                )

            # make sure there are no undefined fields
            defined_field_names = self.get_defined_fields(
                defining_definition.package, defining_definition.name
            )
            item_field_names = [field for field in item.keys()]
            for item_field_name in item_field_names:
                if item_field_name not in defined_field_names:
                    raise LanguageError(
                        f"Found undefined field name '{item_field_name}' when expecting {defined_field_names} as defined in {defining_definition.name}",
                        self.get_location_str(item_field_name, lexemes),
                    )
            subfields = self.populate_sub_fields(subfields, defining_definition, item, lexemes, definition)

            instance.append(
                self.create_object_instance(instance_class, subfields)
            )
        return instance

    def field_instance_creator_not_list(self, lexemes: list, field_value: Any, field_name: str, defining_definition: Definition, instance: list, instance_class: Type, definition: Definition) -> list:
        """
        Method to create instance fields from non-list type fields.

        Args:
            lexemes (List[Lexeme]): A list of definition Lexemes.
            field_value (Any): Value stored in the field.
            field_name (str): The name of the field.
            defining_definition (Definition): Definition of the field type.
            instance (list[Any]): The instance that will contain the created field
            instance_class (Type): The instance type class
            definition (Definition): Definition containing the field being checked.

        Returns:
            The updated instance containing the created field
        """
        defined_field_names = self.get_defined_fields(
            defining_definition.package, defining_definition.name
        )
        item_field_names = [field for field in field_value.keys()]
        for item_field_name in item_field_names:
            if item_field_name not in defined_field_names:
                raise LanguageError(
                    f"Found undefined field name '{item_field_name}' when expecting {defined_field_names} as defined in {defining_definition.name}",
                    self.get_location_str(item_field_name, lexemes),
                )

        subfields = {}
        if "fields" not in defining_definition.structure["schema"]:
            raise LanguageError(
                f"Schema '{defining_definition.name}' does not contain any fields.",
                self.get_location_str(field_name, lexemes),
            )
        subfields = self.populate_sub_fields(subfields, defining_definition, field_value, lexemes, definition)
        instance = self.create_object_instance(instance_class, subfields)
        return instance

    def field_instance_check(self, is_list: bool, field_value: Any, field_name: str, is_required: bool, lexemes: list[Lexeme], defining_definition: Definition, instance_class: Type, definition: Definition) -> list:
        """
        Method used to check if a field is a list, and to call the corresponding field instance creator method.

        Args:
            is_list (bool): Boolean value determining if the field type is a list.
            field_value (Any): Value stored in the field.
            field_name (str): The name of the field.
            lexemes (List[Lexeme]): A list of definition Lexemes.
            defining_definition (Definition): Definition of the field type.
            instance_class (Type): The class type of the instance being created.
            definition (Definition): Definition containing the field being checked.

        Returns:
            The created instance list.
        """
        if is_list:
            instance = []
            if not field_value:
                if is_required:
                    raise LanguageError(
                        message=f"Missing required field {field_name}",
                        location=self.get_location_str(definition.name, lexemes)
                    )
            else:
                if not isinstance(field_value, list):
                    if is_required:
                        raise LanguageError(
                            f"Invalid parsed value for field '{field_name}'.  Expected type 'list', but found '{type(field_value)}'.  Value = {field_value}",
                            self.get_location_str(field_name, lexemes),
                        )
                    else:
                        return instance
                instance = self.field_instance_creator_list(lexemes, field_value, field_name, defining_definition, instance, instance_class, definition)
        else:
            instance = None
            if not field_value:
                if is_required:
                    raise LanguageError(
                        f"Missing required field {field_name}",
                        self.get_location_str(field_name, lexemes),
                    )
                else:
                    return instance
            if not isinstance(field_value, dict):
                # this is a complex type defined by a schema, with a field value so it should be a dict
                raise LanguageError(
                    f"Invalid parsed value for field '{field_name}'.  Expected type 'dict', but found '{type(field_value)}'.  Value = {field_value}",
                    self.get_location_str(field_name, lexemes),
                )

            # make sure there are no undefined fields
            instance = self.field_instance_creator_not_list(lexemes, field_value, field_name, defining_definition, instance, instance_class, definition)
        return instance

    def primitive_field_value_check(self, is_list: bool, field_value: Any, field_name: str, is_required: bool, lexemes: list, defining_definition: Definition, definition: Definition) -> Any:
        """
        Method used to ensure primitive type field definitions have valid values.

        Args:
            is_list (bool): Boolean value determining if the field type is a list.
            field_value (Any): Value stored in the field.
            field_name (str): The name of the field.
            is_required (bool): Boolean value determining if the field is required.
            lexemes (List[Lexeme]): A list of definition Lexemes.
            defining_definition (Definition): Definition defining the field type.
            definition (Definition): Definition containing the field being checked.

        Returns:
            The content of the primitive type field value.
        """
        python_type = defining_definition.structure["primitive"]["python_type"]
        if not field_value:
            if is_required:
                raise LanguageError(message=f"Missing required field {field_name}.", location=self.get_location_str(definition.name, lexemes))
            if is_list:
                field_value = []
        elif is_list:
            for item in field_value:
                if not isinstance(item, eval(python_type)):
                    raise LanguageError(
                        message=f"Invalid value for field '{field_name}'.  Expected type '{python_type}', but found '{type(item)}'",
                        location=self.get_location_str(field_value, lexemes),
                    )
        else:
            if "Any" != python_type:
                if not isinstance(field_value, eval(python_type)):
                    raise LanguageError(
                        f"Invalid value for field '{field_name}'.  Expected type '{python_type}', but found '{type(field_value)}'",
                        self.get_location_str(field_value, lexemes),
                    )
        return field_value

    def enum_field_list_check(self, field_value: Any, field_name: str, lexemes: list, defining_definition: Definition, enum_class: str) -> list:
        """
        Method used to ensure each item in a list of enum fields have valid values.

        Args:
            field_value (Any): Value stored in the field.
            field_name (str): The name of the field.
            lexemes (List[Lexeme]): A list of definition Lexemes.
            defining_definition (Definition): Definition containing the field being checked.
            enum_class (str): The name of the enums defining definition

        Returns:
            The list of valid enum field values.
        """
        result = []
        for item in field_value:
            if not isinstance(item, str):
                raise LanguageError(
                    f"Invalid value for field '{field_name}'.  Expected type 'str', but found '{type(item)}'",
                    self.get_location_str(field_name, lexemes),
                )
            try:
                result.append(getattr(enum_class, item))
            except ValueError:
                raise LanguageError(
                    f"{item} is not a valid value for enum {defining_definition.name}",
                    self.get_location_str(item, lexemes),
                )
        return result

    def enum_field_value_check(self, is_list: bool, field_value: Any, field_name: str, lexemes: list, defining_definition: Definition) -> list:
        """
        Method used to ensure enum type field definitions have valid values.

        Args:
            is_list (bool): Boolean value determining if the field type is a list.
            field_value (Any): Value stored in the field.
            field_name (str): The name of the field.
            lexemes (List[Lexeme]): A list of definition Lexemes.
            defining_definition (Definition): Definition containing the field being checked.

        Returns:
            The list of valid enum field values.
        """
        try:
            enum_class = self.context.context_instance.fully_qualified_name_to_class[
                defining_definition.get_fully_qualified_name()
            ]
        except LanguageError as e:
            raise LanguageError(
                f"Failed to create Enum instance_class for {defining_definition.name}: {e.message}",
                self.get_location_str(defining_definition.name, defining_definition.lexemes),
            )
        if not enum_class:
            enum_class = self.create_enum_class(defining_definition)
        if is_list:
            return self.enum_field_list_check(field_value, field_name, lexemes, defining_definition, enum_class)
        else:
            if not field_value:
                return [None]
            try:
                return self.context.create_aac_enum(
                    defining_definition.get_fully_qualified_name(), field_value
                )
            except ValueError:
                raise LanguageError(
                    f"{field_value} is not a valid value for enum {defining_definition.name}",
                    self.get_location_str(field_value, lexemes),
                )

    def schema_field_value_check(self, is_list: bool, field_value: Any, field_name: str, is_required: bool, lexemes: list, defining_definition: Definition, definition: Definition) -> list:
        """
        Method used to ensure schema type field definitions have valid values.

        Args:
            is_list (bool): Boolean value determining if the field type is a list.
            field_value (Any): Value stored in the field.
            field_name (str): The name of the field.
            lexemes (List[Lexeme]): A list of definition Lexemes.
            defining_definition (Definition): Definition of the field type.
            definition (Definition): Definition containing the field being checked.

        Returns:
            The List of valid schema defined field values.
        """
        try:
            field_fully_qualified_name = (
                defining_definition.get_fully_qualified_name()
            )
        except LanguageError as e:
            raise LanguageError(
                f"Failed to create fully qualified name for definition {defining_definition.name}: {e.message}",
                self.get_location_str(defining_definition.name, defining_definition.lexemes),
            )
        instance_class = self.context.context_instance.fully_qualified_name_to_class[
            field_fully_qualified_name
        ]
        if not instance_class:
            if defining_definition.get_root_key() == "schema":
                instance_class = self.create_schema_class(defining_definition)
            else:
                raise LanguageError(
                    f"Unable to process AaC definition of type {field_fully_qualified_name} with root {defining_definition.get_root_key()}",
                    self.get_location_str(field_name, lexemes),
                )
        instance = self.field_instance_check(is_list, field_value, field_name, is_required, lexemes, defining_definition, instance_class, definition)

        return instance

    def create_field_instance(
        self,
        field_name: str,
        field_type: str,
        is_required: bool,
        field_value: Any,
        lexemes: list[Lexeme],
        definition: Definition
    ) -> Any:
        """
        Adds an entry to the instance attribute of a definition for the given field.

        Args:
            field_name (str): Name of the field.
            field_type (str): Type of the field.
            is_required (bool): Contents of the is_required field for the specified field.
            field_value (Any): The value for the specified field.
            lexemes (list[Lexeme]): A list of definition Lexemes.
            definition (Definition): Definition containing the field instance being created

        Returns:
            The instance field value.
        """
        is_list = False
        clean_field_type = field_type
        if field_type.endswith("[]"):
            is_list = True
            clean_field_type = field_type[:-2]
        if "(" in clean_field_type:
            clean_field_type = clean_field_type[: clean_field_type.find("(")]

        # now get the defining definition from the clean_field_type
        defining_definitions = self.find_definitions_by_name(clean_field_type)
        if not defining_definitions or len(defining_definitions) == 0:
            raise LanguageError(
                f"Could not find definition for '{clean_field_type}'.",
                self.get_location_str(field_type, lexemes),
            )
        elif len(defining_definitions) > 1:
            raise LanguageError(
                f"Found multiple definitions for '{clean_field_type}'.",
                self.get_location_str(field_type, lexemes),
            )
        defining_definition = defining_definitions[0]

        if defining_definition.get_root_key() == "primitive":
            # this is a primitive, so ensure the parsed value aligns with the type and return it
            return self.primitive_field_value_check(is_list, field_value, field_name, is_required, lexemes, defining_definition, definition)

        elif defining_definition.get_root_key() == "enum":
            # this is an enum, so ensure the parsed value aligns with the type and return it
            return self.enum_field_value_check(is_list, field_value, field_name, lexemes, defining_definition)

        else:  # this isn't a primitive and isn't an enum, so it must be a schema
            return self.schema_field_value_check(is_list, field_value, field_name, is_required, lexemes, defining_definition, definition)

    def create_definition_instance(self, definition: Definition) -> Any:
        """
        Populates the instance field of a given definition.

        Args:
            definition (Definition): The given definition being populated.

        Returns:
            The populated instance for the given definition.
        """
        instance = None

        defining_definition = None
        for item in self.context.get_definitions() + self.parsed_definitions:
            if item.get_root_key() == "schema":
                if "root" in item.structure["schema"]:
                    if (
                        definition.get_root_key()
                        == item.structure["schema"]["root"]
                    ):
                        defining_definition = item

        if not defining_definition:
            raise LanguageError(
                f"Could not find definition for {definition.name} with root {definition.get_root_key()}",
                self.get_location_str(definition.get_root_key(), definition.lexemes),
            )

        if defining_definition.get_root_key() == "schema":
            # since schemas are how we define "all the things" the root key of the defining definition should be 'schema'
            self.create_schema_class(defining_definition)
        else:
            raise LanguageError(
                f"Definition for root key '{defining_definition.get_root_key()}' is not a Schema.",
                self.get_location_str(definition.get_root_key(), definition.lexemes),
            )

        instance = self.create_field_instance(
            "root",
            defining_definition.name,
            True,
            definition.structure[definition.get_root_key()],
            definition.lexemes,
            definition,
        )

        definition.instance = instance
        return instance

    def set_qualified_name(self, definition: Definition):
        """
        Method to set the fully qualified name for a definition.

        Args:
            definition (Definition): The definition whose fully qualified name will be set.
        """
        if definition.get_root_key() == "primitive":
            self.primitive_name_to_py_type[definition.name] = definition.structure["primitive"]["python_type"]
        if "package" in definition.structure[definition.get_root_key()]:
            try:
                fully_qualified_name = definition.get_fully_qualified_name()
            except LanguageError as e:
                raise LanguageError(
                    f"Failed to create fully qualified name for the definition {definition.name}: {e.message}",
                    self.get_location_str(definition.name, definition.lexemes),
                )

            # This is so requirements specifically do not get overwritten.  Although other definition types may still get overwritten, so we may need to find a better solution eventually.
            if definition.get_root_key() == "req":
                req_id = definition.structure["req"]["id"]
                fully_qualified_name = f"{fully_qualified_name}_{req_id}"
            self.fully_qualified_name_to_definition[fully_qualified_name] = definition

    # Start the load_definition function code here
    def load_definitions(
        self, context, parsed_definitions: list[Definition]
    ) -> list[Definition]:
        """
        Loads the given definitions into the context and populates the instance with a python object.

        Args:
            context (LanguageContext): An instance of the active LanguageContext.
            parsed_definitions: (list[Definition]): The parsed contents of a definition file.

        Returns:
            The parsed definitions to load into the LanguageContext.
        """

        # Maintainer note:  Yes, this function is a bit of a monster...sorry about that.
        # I wanted to keep all this stuff together because if it all works out this should be the
        # only place where we have to deal with navigating the structure of the definitions and
        # not using the python objects.  In order for this to work, any changes in here should
        # avoid the use of he definition instance...other than actually creating it.
        schema_defs_by_root = {}
        self.context = context
        self.parsed_definitions = parsed_definitions
        self.primitive_name_to_py_type = {}
        self.fully_qualified_name_to_definition = {}
        for definition in self.parsed_definitions + self.context.get_definitions():
            self.set_qualified_name(definition)
        for definition in self.context.get_definitions():
            if definition.get_root_key() == "schema":
                if definition.instance.root:
                    schema_defs_by_root[definition.instance.root] = definition

        result = []
        for definition in self.parsed_definitions:
            # create and register the instance
            self.create_definition_instance(definition)
            definition.source.is_loaded_in_context = True
            result.append(definition)
            self.context.context_instance.definitions.add(definition)
            fully_qualified_name = f"{definition.package}.{definition.name}"

            # This is so requirements specifically do not get overwritten.  Although other definition types may still get overwritten, so we may need to find a better solution eventually.
            if definition.get_root_key() == "req":
                req_id = definition.structure["req"]["id"]
                fully_qualified_name = f"{fully_qualified_name}_{req_id}"
            self.context.context_instance.fully_qualified_name_to_definition[
                fully_qualified_name
            ] = definition
        return result



"""File metadata and management class for AaC."""
from attr import attrib, attrs, validators


@attrs(hash=False, eq=False)
class AaCFile:
    """
    An Architecture-as-Code file containing AaC definitions.

    Attributes:
        uri (str): The file's URI path
        is_user_editable (bool): True if the file can be edited by the user, otherwise false
        is_loaded_in_context (bool): True if the file is currently loaded into the context, otherwise false
    """
    uri: str = attrib(validator=validators.instance_of(str))
    is_user_editable: bool = attrib(validator=validators.instance_of(bool))
    is_loaded_in_context: bool = attrib(validator=validators.instance_of(bool))

    def __hash__(self):
        """Hash function for the class."""
        return hash(self.uri)

    def __eq__(self, obj):
        """Equals function for the class."""
        return isinstance(obj, AaCFile) and obj.uri == self.uri



"""The AaC Exclusive fields plugin implementation module."""
# NOTE: It is safe to edit this file.
# This file is only initially generated by aac gen-plugin, and it won't be overwritten if the file already exists.

# There may be some unused imports depending on the definition of the plugin...but that's ok
from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
    ExecutionMessage,
    MessageLevel,
)
from aac.context.definition import Definition
from typing import Any


plugin_name = "Exclusive Fields"


def mutually_exclusive_fields(
    instance: Any, definition: Definition, defining_schema, fields: list[str]
) -> ExecutionResult:
    """Business logic for the Mutually exclusive fields constraint."""

    num_present = 0
    for field in fields:
        instance_fields = vars(instance)
        if field in instance_fields:
            field_value = getattr(instance, field)
            if field_value is None:  # handle primitives and schema defined types
                continue
            elif (
                isinstance(field_value, list) and len(field_value) == 0
            ):  # handle lists
                continue
            elif (
                isinstance(field_value, dict) and len(field_value) == 0
            ):  # this is for the any type
                continue
            else:
                num_present += 1  # we found a real instance, so count it

    if num_present == 0:
        error_msg = ExecutionMessage(
            f"None of the following fields were present: {fields}",
            MessageLevel.ERROR,
            definition.source.uri,
            None,
        )
        return ExecutionResult(
            plugin_name,
            "Mutually exclusive fields",
            ExecutionStatus.GENERAL_FAILURE,
            [error_msg],
        )

    if num_present > 1:
        error_msg = ExecutionMessage(
            f"More than one of the following fields were present: {fields}",
            MessageLevel.ERROR,
            definition.source.uri,
            None,
        )
        return ExecutionResult(
            plugin_name,
            "Mutually exclusive fields",
            ExecutionStatus.GENERAL_FAILURE,
            [error_msg],
        )

    return ExecutionResult(
        plugin_name, "Mutually exclusive fields", ExecutionStatus.SUCCESS, []
    )



from unittest import TestCase

from aac.context.definition import Definition
from aac.context.language_context import LanguageContext


class TestDefinition(TestCase):

    def test_get_root_key(self):
        context = LanguageContext()
        definition = context.get_definitions_by_name("Schema")
        self.assertEqual(definition[0].get_root_key(), "schema")

        definition = context.get_definitions_by_name("Field")
        self.assertEqual(definition[0].get_root_key(), "schema")

        definition = context.get_definitions_by_name("string")
        self.assertEqual(definition[0].get_root_key(), "primitive")

        definition = context.get_definitions_by_name("abstract")
        self.assertEqual(definition[0].get_root_key(), "modifier")

    def test_is_import(self):
        context = LanguageContext()
        definition = context.parse_and_load(IMPORT_DEFINITION)[0]
        self.assertTrue(definition.is_import())
        definition = context.get_definitions_by_name("Schema")[0]
        self.assertFalse(definition.is_import())

    def test_get_python_module_name(self):
        context = LanguageContext()
        definition = context.get_definitions_by_name("Schema")[0]
        module_name = definition.get_python_module_name()
        self.assertEqual(module_name, "aac.lang")

    def test_get_python_class_name(self):
        context = LanguageContext()
        definition = context.parse_and_load(DEFINITION_VALID)[0]
        class_name = definition.get_python_class_name()
        self.assertEqual(class_name, "ModelName")

    def test_get_fully_qualified_name(self):
        context = LanguageContext()
        definition = context.get_definitions_by_name("Schema")[0]
        name = definition.get_fully_qualified_name()
        self.assertEqual(name, "aac.lang.Schema")

    def test_to_yaml(self):
        context = LanguageContext()
        definition = context.get_definitions_by_name("Schema")
        yaml = definition[0].to_yaml()
        self.assertIn("schema:", yaml)
        self.assertIn("name: Schema", yaml)

    def test_definition_fail_incorrect_types(self):
        context = LanguageContext()
        schema_definition = context.get_definitions_by_name("Schema")[0]
        with self.assertRaises(TypeError):
            # This is probably the worst way to make a definition, it should normally be handled by Language Context
            definition = Definition(
                name=123,
                package=schema_definition.package,
                content=schema_definition.content,
                source=schema_definition.source,
                structure=schema_definition.structure,
            )
        with self.assertRaises(TypeError):
            definition = Definition(            # noqa: F841
                name=schema_definition.name,
                package=123,
                content=schema_definition.content,
                source=schema_definition.source,
                structure=schema_definition.structure,
            )
        with self.assertRaises(TypeError):
            definition = Definition(            # noqa: F841
                name=schema_definition.name,
                package=schema_definition.package,
                content=123,
                source=schema_definition.source,
                structure=schema_definition.structure,
            )

    def test_get_python_module_name_default(self):
        context = LanguageContext()
        definition = context.parse_and_load(DEFINITION_VALID)[0]
        module_name = definition.get_python_module_name()
        self.assertEqual(module_name, "default")


IMPORT_DEFINITION = """
import:
  files:
    - ./structures.yaml
    - ../alarm_clock/structures.yaml
"""

DEFINITION_VALID = """
model:
    name: Model Name
"""

DEFINITION_INVALID_ROOT_KEY = """
notaroot:
    name: notaroot
"""



"""Module to comb directories and identify AaC files."""
import logging
import os

from aac.in_out.constants import YAML_DOCUMENT_EXTENSION, AAC_DOCUMENT_EXTENSION
from aac.in_out.parser import parse, ParserError
from aac.in_out.files.aac_file import AaCFile


def find_aac_files(root_directory_to_search: str) -> list[AaCFile]:
    """Return a list of paths to AaC files in or under the root directory."""
    aac_files = []

    root_dir_abs_path = os.path.abspath(root_directory_to_search)
    if os.path.isdir(root_dir_abs_path):
        directory_results = os.walk(root_dir_abs_path)
        for directory_path, _, filenames in directory_results:
            for filename in filenames:
                filepath = os.path.join(directory_path, filename)
                if is_aac_file(filepath):
                    aac_files.append(AaCFile(filepath, True, False))
    else:
        logging.error(f"Root path '{root_dir_abs_path}' for AaC file discovery is not a directory.")

    return aac_files


def is_aac_file(filepath: str) -> bool:
    """Test if a target file is considered a valid AaC file."""
    _, file_ext = os.path.splitext(filepath)

    is_valid_aac_file = False
    if file_ext == YAML_DOCUMENT_EXTENSION or file_ext == AAC_DOCUMENT_EXTENSION:
        try:
            parse(filepath)
        except ParserError as error:
            logging.error(f"File '{filepath}' is not a valid AaC file. Reason: {error.errors}")
        except Exception as error:
            logging.error(f"File '{filepath}' could not be parsed. Reason: {error}")
        else:
            is_valid_aac_file = True

    return is_valid_aac_file



"""Parser Metadata class use to manage and identify the location of definitions in the source YAML."""

from attr import attrib, attrs, validators


@attrs
class SourceLocation:
    """The position and span of an AaC structure in the YAML source.

    Attributes:
        line (int): The line number on which the object was found.
        column (int): The character position at which the object was found.
        position (int): The position relative to the start of the file where the object was found.
        span (int): The number of characters occupied by the object relative to `position`.
    """

    line: int = attrib(validator=validators.instance_of(int))
    column: int = attrib(validator=validators.instance_of(int))
    position: int = attrib(validator=validators.instance_of(int))
    span: int = attrib(validator=validators.instance_of(int))

    def to_tuple(self) -> tuple[int, int, int, int]:
        """
        Return a representation of the location as a tuple.

        Returns:
            An (int, int, int, int) tuple consisting of the line number, column, character start, and the span.
        """
        return self.line, self.column, self.position, self.span



"""The AaC Unique Root Keys plugin implementation module."""
# NOTE: It is safe to edit this file.
# This file is only initially generated by aac gen-plugin, and it won't be overwritten if the file already exists.

from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
    ExecutionMessage,
    MessageLevel,
)
from aac.context.language_context import LanguageContext

plugin_name = "Unique Root Keys"


def root_key_names_are_unique(context: LanguageContext) -> ExecutionResult:
    """Business logic for the Root key names are unique constraint."""

    status = ExecutionStatus.SUCCESS
    messages: list[ExecutionMessage] = []

    root_keys = []
    for definition in context.get_definitions():
        if definition.get_root_key() == "schema":
            if definition.instance.root:
                if definition.instance.root not in root_keys:
                    root_keys.append(definition.instance.root)
                else:
                    status = ExecutionStatus.CONSTRAINT_FAILURE
                    error_msg = ExecutionMessage(
                        f"Root key {definition.instance.root} is not unique.",
                        MessageLevel.ERROR,
                        definition.source.uri,
                        None,
                    )
                    messages.append(error_msg)

    return ExecutionResult(plugin_name, "Root key names are unique", status, messages)



"""The AaC Check AaC plugin implementation module."""
# NOTE: It is safe to edit this file.
# This file is only initially generated by aac gen-plugin, and it won't be overwritten if the file already exists.

from typing import Callable, Any
from aac.context.language_context import LanguageContext
from aac.context.definition import Definition
from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
    ExecutionMessage,
    MessageLevel,
)
from aac.context.language_error import LanguageError

plugin_name = "Check AaC"


def check(aac_file: str, fail_on_warn: bool, verbose: bool) -> ExecutionResult:  # noqa: C901
    """Business logic for the check command.

    Args:
        aac_file (str):         The AaC file being processed
        fail_on_warn (bool):    Flag to fail when warnings are discovered
        verbose (bool):         Flag for verbose mode. When true add success messages as encountered.

    Returns:
        ExecutionResult:        Method result containing: plugin_name ("Check AaC"), "check", status, message
                                including results from lower level helper methods

    Raises:
        LanguageError: Passed up LanguageError from get_defining_schema_for_root
    """

    constraint_results: dict[str, list[ExecutionResult]] = {}

    context: LanguageContext = LanguageContext()

    # collect all constraints for easy access
    all_constraints_by_name: dict[str, Callable] = {}
    for runner in context.get_plugin_runners():
        for name, callback in runner.constraint_to_callback.items():
            all_constraints_by_name[name] = callback

    # we'll need to recurse our way through the schema to check all the constraints
    # so we'll create a couple functions to help us navigate the way
    def check_primitive_constraint(
        field,
        source_definition: Definition,
        value_to_check: Any,
        primitive_declaration: str,
        defining_primitive,
    ):
        """Helper method that runs all the constraints for a given primitive.

        Args:
            field:                          The field being checked
            source_definition (Definition): Source of the check_me field that we are evaluating
            value_to_check (Any):           The field value being checked
            primitive_declaration (str):    The declaration of the primitive
            defining_primitive:             The defining primitive constraints

        Returns:
            (at check level) ExecutionResult: The result of the checks in this helper method
        """

        # Check the value_to_check against the defining_primitive
        defining_primitive_instance = defining_primitive
        for constraint_assignment in defining_primitive_instance.constraints:
            constraint_name = constraint_assignment.name
            constraint_args = constraint_assignment.arguments
            callback = all_constraints_by_name[constraint_name]
            # This location code feels like a hack!  Is there a better way?
            locations = [
                lexeme.location
                for lexeme in source_definition.lexemes
                if lexeme.value == field.name
            ]
            location = None
            if len(locations) > 0:
                location = locations[0]

            result: ExecutionResult = callback(
                value_to_check,
                primitive_declaration,
                constraint_args,
                source_definition.source,
                location,
            )
            if constraint_name not in constraint_results:
                constraint_results[constraint_name] = []
            constraint_results[constraint_name].append(result)

    def check_schema_constraint(
        source_definition: Definition, check_me: Any, check_against
    ):
        """Helper method that runs all the constraints for a given schema.

        Args:
            source_definition (Definition): Source of the check_me field that we are evaluating
            check_me (Any):                 The field being checked
            check_against:                  The schema we are comparing the check_me field against

        Returns:
            (at check level) ExecutionResult: The result of the checks in this helper method

        Raises:
            LanguageError: If unique schema definition for field type not found for field name
            LanguageError: If value of field name was something other than a list
        """

        # make sure we've got a schema
        context = LanguageContext()
        if not context.is_aac_instance(check_against, "aac.lang.Schema"):
            return
        # collect applicable constraints
        schema_constraints = []
        for runner in context.get_plugin_runners():
            plugin = runner.plugin_definition.instance
            for constraint in plugin.schema_constraints:
                if constraint.universal:
                    schema_constraints.append(
                        context.create_aac_object(
                            "SchemaConstraintAssignment",
                            {"name": constraint.name, "arguments": []},
                        )
                    )
        if check_against.constraints:
            for constraint_assignment in check_against.constraints:
                schema_constraints.append(constraint_assignment)

        # Check the check_me against constraints in the defining_schema
        for constraint_assignment in schema_constraints:
            constraint_name = constraint_assignment.name
            constraint_args = constraint_assignment.arguments
            callback = all_constraints_by_name[constraint_name]
            result: ExecutionResult = callback(
                check_me, source_definition, check_against, constraint_args
            )

            if constraint_name not in constraint_results:
                constraint_results[constraint_name] = []
            constraint_results[constraint_name].append(result)

        # loop through the fields on the check_against schema
        for field in check_against.fields:
            # only check the field if it is present
            if not hasattr(check_me, field.name):
                continue

            # get the name of the schema that defines the field, special handling for arrays and references
            type_name = field.type
            is_list = False
            # if type name ends with "[]", remove the brackets and set is_list to True
            if field.type.endswith("[]"):
                type_name = field.type[: -2]
                is_list = True
            # if type name has parameters in parens, remove them
            if type_name.find("(") > -1:
                type_name = type_name[: type_name.find("(")]

            # get the definition that defines the field
            field_defining_schema = context.get_definitions_by_name(type_name)

            if len(field_defining_schema) != 1:
                # Question: should we convert this to a Constraint Failure?
                raise LanguageError(
                    f"Could not find unique schema definition for field type {field.type} with name {field.name}",
                    source_definition.source.uri
                )

            if field_defining_schema[0].get_root_key() == "primitive":
                # if the field is a primitive, run the primitive constraints

                if is_list:
                    # if the field is a list, check each item in the list

                    if type(getattr(check_me, field.name)) != list:
                        raise LanguageError(
                            f"Value of '{field.name}' was expected to be list, but was '{type(getattr(check_me, field.name))}'",
                            source_definition.source.uri
                        )

                    for item in getattr(check_me, field.name):
                        value_to_check = item
                        if value_to_check is not None:

                            check_primitive_constraint(
                                field,
                                source_definition,
                                item,
                                field.type[:-2],
                                field_defining_schema[0].instance,
                            )
                else:
                    value_to_check = getattr(check_me, field.name)
                    if value_to_check is not None:
                        check_primitive_constraint(
                            field,
                            source_definition,
                            value_to_check,
                            field.type,
                            field_defining_schema[0].instance,
                        )
            else:
                # if the field is a schema, run the schema constraints
                if is_list:
                    # if the field is a list, check each item in the list
                    for item in getattr(check_me, field.name):

                        check_schema_constraint(
                            source_definition, item, field_defining_schema[0].instance
                        )
                else:

                    check_schema_constraint(
                        source_definition,
                        getattr(check_me, field.name),
                        field_defining_schema[0].instance,
                    )

    # now that the helper functions are in place, let's run the constraints on the aac_file

    # FIX ME: This call to parse_and_load can throw LanguageError and ParserError exceptions which are not being handled here
    # FIX ME: They should be handled, their messages added to the message list being constructed in this method (complete with source and location info)
    # FIX ME: so they can be passed along. This was discovered during the expansion of test_check_aac.py and those tests would need to be updated.
    definitions_to_check = context.parse_and_load(aac_file)

    # First run all context constraint checks
    # Context constraints are "language constraints" and are not tied to a specific schema
    # You can think of these as "invariants", so they must always be satisfied
    for plugin in context.get_definitions_by_root("plugin"):
        # we want to check contest constraints, but not the ones that are defined in the aac_file we're checking to avoid gen-plugin circular logic
        for context_constraint in plugin.instance.context_constraints:
            if context_constraint.name not in [
                definition.name for definition in definitions_to_check
            ]:
                if context_constraint.name in all_constraints_by_name.keys():
                    callback = all_constraints_by_name[context_constraint.name]
                    result: ExecutionResult = callback(context)
                    if context_constraint.name not in constraint_results:
                        constraint_results[context_constraint.name] = []
                    constraint_results[context_constraint.name].append(result)

    for check_me in definitions_to_check:
        try:
            defining_schema = context.get_defining_schema_for_root(check_me.get_root_key())
        except LanguageError as e:
            raise LanguageError(e.message, check_me.source.uri)
        check_schema_constraint(check_me, check_me.instance, defining_schema.instance)

    # loop through all the constraint results and see if any of them failed
    messages = []
    status = ExecutionStatus.SUCCESS
    for name, results in constraint_results.items():
        for result in results:
            if result.is_success():
                # if the result is a success, add the messages to the list if we're in verbose mode
                # because these should only be info messages
                if verbose:
                    messages.extend(result.messages)
            elif result.status_code == ExecutionStatus.CONSTRAINT_WARNING:
                # if the result is a warning, add the messages to the list and fail the check if fail_on_warn is true
                if fail_on_warn:
                    status = ExecutionStatus.CONSTRAINT_FAILURE
                messages.extend(result.messages)
            else:
                # Any failure (including a constraint failure) is handled the same way
                messages.extend(result.messages)
                # don't change the status if already a failure
                if status != ExecutionStatus.CONSTRAINT_FAILURE:
                    status = result.status_code

    # after going through all the constraint results, if we're still successful, add a success message
    if verbose:
        for check_me in definitions_to_check:
            messages.append(
                ExecutionMessage(
                    f"Check {check_me.source.uri} - {check_me.name} was successful.",
                    level=MessageLevel.DEBUG,
                    source=check_me.source.uri,
                    location=None,
                )
            )
    if status == ExecutionStatus.SUCCESS:
        messages.append(
            ExecutionMessage(
                message="All AaC constraint checks were successful.",
                level=MessageLevel.INFO,
                source=aac_file,
                location=None,
            )
        )

    return ExecutionResult(plugin_name, "check", status, messages)



"""AaC Parser and related functions submodule."""

from aac.in_out.parser._parse_source import parse
from aac.in_out.parser._parser_error import ParserError
from aac.in_out.parser._yaml import scan_yaml, parse_yaml
from aac.in_out.parser._cache_manager import get_cache, reset_cache

__all__ = (
    ParserError.__name__,
    parse.__name__,
    scan_yaml.__name__,
    parse_yaml.__name__,
    get_cache.__name__,
    reset_cache.__name__,
)



from unittest import TestCase
from typing import Tuple
from click.testing import CliRunner
from aac.execute.command_line import cli, initialize_cli
from aac.execute.aac_execution_result import ExecutionStatus


from aac.plugins.version.version_impl import version


class TestVersion(TestCase):

    def test_version(self):
        version_result = version()
        # make sure the command was successful
        self.assertEqual(version_result.status_code, ExecutionStatus.SUCCESS)

        # get the reported version string
        message = version_result.messages[0]

        # make sure the version conforms to semantic versioning
        version_values = message.message.split(".")
        self.assertEqual(len(version_values), 3, "Version should have 3 values separated by . to follow semantic versioning standards.")
        for value in version_values:
            self.assertTrue(value.isdigit(), msg=f"Each token of the version should be a digit.  Received {value}")

    def run_version_cli_command_with_args(self, args: list[str]) -> Tuple[int, str]:
        """Utility function to invoke the CLI command with the given arguments."""
        initialize_cli()
        runner = CliRunner()
        result = runner.invoke(cli, ["version"] + args)
        exit_code = result.exit_code
        output_message = result.output.strip().replace("\x1b[0m", "")
        return exit_code, output_message

    def test_cli_version(self):
        args = []

        exit_code, output_message = self.run_version_cli_command_with_args(args)

        self.assertEqual(0, exit_code)  # asserts the command ran successfully
        self.assertGreater(len(output_message), 0)  # asserts the command produced output
        version_values = output_message.strip().split(".")
        self.assertEqual(len(version_values), 3, "Version should have 3 values separated by . to follow semantic versioning standards.")
        for value in version_values:
            self.assertTrue(value.isdigit(), msg=f"Each token of the version should be a digit.  Received {value}")



"""Provide access to plugins and plugin data."""

from importlib import import_module
from pkgutil import iter_modules
from types import ModuleType
from pluggy import PluginManager
from typing import List

from aac.execute import PLUGIN_PROJECT_NAME, hookspecs


def get_plugin_manager() -> PluginManager:
    """
    Get the plugin manager and automatically register core plugins.

    Returns:
        The plugin manager.
    """
    plugin_manager = PluginManager(PLUGIN_PROJECT_NAME)
    plugin_manager.add_hookspecs(hookspecs)
    plugin_manager.load_setuptools_entrypoints(PLUGIN_PROJECT_NAME)

    plugins = [
        *register_plugins_in_package("aac.plugins"),
    ]

    for plugin in plugins:
        plugin_manager.register(plugin)

    return plugin_manager


def register_plugins_in_package(package: str) -> List[ModuleType]:
    """
    Register all the plugins in the specified package.

    Note, this function depends on the ability to import package and its direct child packages.

    Args:
        package (str): The package in which to find plugins to be registered.

    Returns:
        A list of top-level plugin modules that define implemented plugins.
    """
    plugins_package = import_module(package)
    return [import_module(f"{package}.{module_name}") for _, module_name, _ in iter_modules(plugins_package.__path__)]



from unittest import TestCase

from aac.execute.plugin_manager import get_plugin_manager


class TestPluginManager(TestCase):
    def test_plugin_manager(self):
        plugin_manager = get_plugin_manager()
        plugins = plugin_manager.get_plugins()
        plugin = next(iter(plugins))
        self.assertTrue(str(plugin), msg=''.join(str(plugins)))

        plugin_manager.unregister(plugin)
        plugins = plugin_manager.get_plugins()
        self.assertNotIn(str(plugin), ''.join(str(plugins)))

        plugin_manager.register(plugin)
        plugins = plugin_manager.get_plugins()
        self.assertTrue(str(plugin), msg=''.join(str(plugins)))

# More investigations for allowing a fail test with pluggy required.  It currently accepts any type, so a fail test is not clear.



"""The AaC No Ext for Final plugin implementation module."""
# NOTE: It is safe to edit this file.
# This file is only initially generated by aac gen-plugin, and it won't be overwritten if the file already exists.

# There may be some unused imports depending on the definition of the plugin...but that's ok
from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
    ExecutionMessage,
    MessageLevel,
)

from aac.context.language_context import LanguageContext
from aac.context.definition import Definition
from typing import Any


plugin_name = "No Extends for Final"


def no_extension_for_final(
    instance: Any, definition: Definition, defining_schema
) -> ExecutionResult:
    """Business logic for the No Extension for Final constraint."""

    status = ExecutionStatus.SUCCESS
    messages: list[ExecutionMessage] = []
    context = LanguageContext()
    if context.is_aac_instance(instance, "aac.lang.Schema"):
        if instance.extends:
            for ext in instance.extends:
                ext_definition = context.get_definitions_by_name(ext.name)
                if len(ext_definition) != 1:
                    status = ExecutionStatus.GENERAL_FAILURE
                    messages.append(
                        ExecutionMessage(
                            f"Cannot resolve unique type for extension {ext.name}.  Found {ext_definition}",
                            MessageLevel.ERROR,
                            definition.source.uri,
                            None,
                        )
                    )
                else:
                    ext_definition = ext_definition[0]
                    if (
                        ext_definition.instance.modifiers
                        and "final" in ext_definition.instance.modifiers
                    ):
                        status = ExecutionStatus.CONSTRAINT_FAILURE
                        messages.append(
                            ExecutionMessage(
                                f"{definition.name} cannot extend {ext.name} because it is final.",
                                MessageLevel.ERROR,
                                ext_definition.source.uri,
                                None,
                            )
                        )

    return ExecutionResult(plugin_name, "No Extension for Final", status, messages)



"""The Architecture-as-Code tool."""

# Run Python3.9 Check before continuing
import sys

if sys.version_info < (3, 9):
    minor = sys.version_info.minor
    major = sys.version_info.major
    print(f"Python version {major}.{minor} is too low; AaC requires at least Python version 3.9 or higher to run.")
    exit(1)

import logging
import os

__version__ = "0.5.11"

__log_file_name__ = os.path.join(os.path.dirname(__file__), "aac.log")

logging.basicConfig(
    format="%(asctime)s::%(pathname)s:%(lineno)d::%(levelname)s::%(message)s",
    filename=__log_file_name__,
    encoding="utf-8",
    level=logging.DEBUG,
    datefmt="%m/%d/%Y %H:%M:%S",
)

__all__ = ("__version__",)



"""The entry-point for the command line interface for the aac tool."""

import sys
from os import linesep
from click import (
    Argument,
    Command,
    Option,
    ParamType,
    Parameter,
    Path,
    UNPROCESSED,
    group,
    secho,
    types,
)

from aac.execute.plugin_runner import AacCommand, AacCommandArgument, PluginRunner
from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
    ExecutionMessage,
    OperationCancelled,
    MessageLevel,
)
from aac.context.definition import Definition
from aac.context.language_error import LanguageError
from aac.context.language_context import LanguageContext
from aac.in_out.parser._parser_error import ParserError

from typing import Callable, Any


@group(context_settings=dict(help_option_names=["-h", "--help"]))
def cli():
    """The Architecture-as-Code (AaC) command line tool."""
    pass


@cli.result_callback()
def output_result(result: ExecutionResult):
    """Output the result of the command."""
    error_occurred = not result.is_success()
    secho(result.get_messages_as_string(), err=error_occurred, color=True)

    if error_occurred:
        sys.exit(result.status_code.value)


def to_click_type(type_name: str) -> ParamType:
    """Convert the named type to a type recognized by Click."""
    if type_name == "file":
        return Path(file_okay=True)
    elif type_name == "directory":
        return Path(dir_okay=True)

    return types.__dict__.get(type_name.upper(), UNPROCESSED)


def to_click_parameter(argument: AacCommandArgument) -> Parameter:
    """Convert an AacCommandArgument to a Click Parameter."""
    names = [argument.name] if isinstance(argument.name, str) else argument.name
    args = dict(
        type=to_click_type(argument.data_type), nargs=1, default=argument.default
    )
    return (
        Option(
            names,
            help=argument.description,
            show_default=True,
            is_flag=argument.data_type == "bool",
            **args,
        )
        if argument.name[0].startswith("-")
        else Argument(names, **args)
    )


def handle_exceptions(plugin_name: str, func: Callable) -> Callable:  # noqa: C901
    """Decorator to catch and handle exceptions in a function."""

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except LanguageError as e:
            return ExecutionResult(
                plugin_name,
                "exception",
                ExecutionStatus.GENERAL_FAILURE,
                [ExecutionMessage(e.message, MessageLevel.ERROR, e.location, None)],
            )
        except OperationCancelled as e:
            return ExecutionResult(
                plugin_name,
                "exception",
                ExecutionStatus.OPERATION_CANCELLED,
                [ExecutionMessage(str(e), MessageLevel.ERROR, None, None)],
            )
        except ParserError as e:
            usr_msg = f"The AaC file '{e.source}' could not be parsed.{linesep}"
            if e.errors:
                usr_msg = f"{usr_msg}The following errors were encountered:{linesep}"
                for err in e.errors:
                    usr_msg += f"  - {err}{linesep}"
            if e.yaml_error:
                usr_msg += f"The following YAML errors were encountered:{linesep}"
                exc = e.yaml_error
                if hasattr(exc, "problem_mark"):
                    if exc.context is not None:
                        usr_msg += f"  Parser Location: {str(exc.problem_mark)} - Problem: {str(exc.problem)} - Context: {str(exc.context)}{linesep}Please correct data and retry."
                    else:
                        usr_msg += f"  Parser Location: {str(exc.problem_mark)} - Problem: {str(exc.problem)}{linesep}Please correct data and retry."
            return ExecutionResult(
                plugin_name,
                "exception",
                ExecutionStatus.PARSER_FAILURE,
                [ExecutionMessage(usr_msg, MessageLevel.ERROR, None, None)],
            )

    return wrapper


def to_click_command(plugin_name: str, command: AacCommand) -> Command:
    """Convert an AacCommand to a Click Command."""

    def is_required_arg(arg):
        if isinstance(arg, list):
            return is_required_arg(arg[0])

        return not arg.name.startswith("-")

    return Command(
        name=command.name,
        callback=handle_exceptions(plugin_name, command.callback),
        params=[to_click_parameter(arg) for arg in command.arguments],
        short_help=command.description,
        no_args_is_help=len([arg for arg in command.arguments if is_required_arg(arg)])
        > 0,
    )


def get_command_arguments(plugin_command: Any, definition: Definition):
    """
    Function to get a list of arguments for a plugin command.

    Args:
        plugin_command (Any): The plugin command with arguments.
        definition (Definition): The plugin definition.

    Returns:
        arguments (list[AacCommandArgument]): A list of command arguments.
    """
    context = LanguageContext()
    arguments: list[AacCommandArgument] = []
    for input in plugin_command.input:
        try:
            arguments.append(
                AacCommandArgument(
                    input.name,
                    input.description,
                    context.get_python_type_from_primitive(input.type),
                    input.default,
                )
            )
        except LanguageError as e:
            raise LanguageError(e.message, definition.source.uri)
    return arguments


def initialize_cli():
    """Initialize the CLI."""
    try:
        active_context = LanguageContext()
    except LanguageError as e:
        if e.location:
            secho(f"{e.message}{linesep}{e.location}", err=True, color=True)
        else:
            secho(f"{e.message}", err=True, color=True)
        sys.exit(1)

    def get_commands() -> list[AacCommand]:
        result: list[AacCommand] = []

        context = LanguageContext()
        for runner in context.get_plugin_runners():
            definition = runner.plugin_definition
            for plugin_command in definition.instance.commands:
                arguments: list[AacCommandArgument] = get_command_arguments(plugin_command, definition)

                result.append(
                    AacCommand(
                        plugin_command.name,
                        plugin_command.help_text,
                        runner.command_to_callback[plugin_command.name],
                        arguments,
                    )
                )
        return result

    runners: list[PluginRunner] = active_context.get_plugin_runners()
    for runner in runners:
        commands = [
            to_click_command(runner.get_plugin_name(), cmd) for cmd in get_commands()
        ]
        for command in commands:
            cli.add_command(command)


# This is the entry point for the CLI
initialize_cli()



"""The AaC Version plugin implementation module."""
# NOTE: It is safe to edit this file.
# This file is only initially generated by aac gen-plugin, and it won't be overwritten if the file already exists.

# There may be some unused imports depending on the definition of the plugin...but that's ok
from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
    ExecutionMessage,
    MessageLevel,
)
from aac import __version__

plugin_name = "Version"


def version() -> ExecutionResult:
    """Business logic for the version command."""

    status = ExecutionStatus.SUCCESS
    messages: list[ExecutionMessage] = []
    version_msg = ExecutionMessage(
        f"{__version__}",
        MessageLevel.INFO,
        None,
        None,
    )
    messages.append(version_msg)

    return ExecutionResult(plugin_name, "version", status, messages)



"""__init__.py module for the Print AaC Definitions plugin."""
# WARNING - DO NOT EDIT - YOUR CHANGES WILL NOT BE PROTECTED.
# This file is auto-generated by the aac gen-plugin and may be overwritten.

from os.path import join, dirname
from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
)
from aac.execute import hookimpl
from aac.context.language_context import LanguageContext
from aac.execute.plugin_runner import PluginRunner
from aac.plugins.print_defs.print_aac_definitions_impl import plugin_name, print_defs

print_aac_definitions_aac_file_name = "print_aac_definitions.aac"


def run_print_defs(core_only) -> ExecutionResult:
    """Print YAML representation of AaC language definitions."""

    result = ExecutionResult(plugin_name, "print-defs", ExecutionStatus.SUCCESS, [])

    print_defs_result = print_defs(core_only)
    if not print_defs_result.is_success():
        return print_defs_result
    else:
        result.add_messages(print_defs_result.messages)

    return result


@hookimpl
def register_plugin() -> None:
    """
    Returns information about the plugin.

    Returns:
        A collection of information about the plugin and what it contributes.
    """

    active_context = LanguageContext()
    print_aac_definitions_aac_file = join(
        dirname(__file__), print_aac_definitions_aac_file_name
    )
    definitions = active_context.parse_and_load(print_aac_definitions_aac_file)

    print_aac_definitions_plugin_definition = [
        definition for definition in definitions if definition.name == plugin_name
    ][0]

    plugin_instance = print_aac_definitions_plugin_definition.instance
    for file_to_load in plugin_instance.definition_sources:
        active_context.parse_and_load(file_to_load)

    plugin_runner = PluginRunner(
        plugin_definition=print_aac_definitions_plugin_definition
    )
    plugin_runner.add_command_callback("print-defs", run_print_defs)

    active_context.register_plugin_runner(plugin_runner)



"""
Parse Architecture-as-Code YAML files.

The AaC parser reads a YAML file, performs validation (if not suppressed) and provides
the caller with a dictionary of the content keyed by the named type.  This allows you
to find a certain type in a model by just looking for that key.
"""
import logging

from copy import deepcopy
from os import path, linesep
from yaml import Mark, Token, StreamStartToken, StreamEndToken, DocumentStartToken

from aac.context.constants import DEFINITION_FIELD_NAME, ROOT_KEY_IMPORT
from aac.context.definition import Definition
from aac.context.lexeme import Lexeme
from aac.context.source_location import SourceLocation
from aac.in_out.constants import DEFAULT_SOURCE_URI, YAML_DOCUMENT_EXTENSION, AAC_DOCUMENT_EXTENSION
from aac.in_out.files.aac_file import AaCFile
from aac.in_out.parser._cache_manager import get_cache
from aac.in_out.paths import sanitize_filesystem_path


YAML_CACHE = get_cache()


def parse(source: str, source_uri: str = DEFAULT_SOURCE_URI) -> list[Definition]:
    """
    Parse the Architecture-as-Code (AaC) definition(s) from the provided source.

    Args:
        source (str): Must be either a file path to an AaC yaml file or a string containing AaC definitions.
        source_uri (Optional[str]): Overrides and sets the source_uri

    Returns:
        A list of Definition objects containing the internal representation of the definition and metadata
        associated with the definition.
    """
    # Linesep provides a quick and cheap filter for filepaths. A valid filepath won't have newlines.
    sanitized_source = source
    is_file = False

    if linesep not in source:
        sanitized_source = sanitize_filesystem_path(source)
        if path.isfile(sanitized_source):
            is_file = True

    return _parse_file(sanitized_source) if is_file else _parse_str(source_uri, source)


def _parse_file(arch_file: str) -> list[Definition]:
    """
    Parse an Architecture-as-Code YAML file and return the definitions in it and its imported files.

    Args:
        arch_file (str): The Architecture-as-Code YAML file to be parsed.

    Returns:
        The AaC definitions extracted from the specified file.
    """
    definitions_list = []
    parsed_files = set()

    def parse_file_contents(file: str):
        if file not in parsed_files:
            file_content = _read_arch_file_content(file)
            parsed_files.add(file)

            if file_content:
                parsed_definitions = _parse_str(file, file_content)
                definitions_list.extend(parsed_definitions)
                [parse_file_contents(imp_file) for imp_file in _get_files_to_import_from_definitions(file, parsed_definitions)]

    parse_file_contents(arch_file)
    return definitions_list


def _parse_str(source: str, model_content: str) -> list[Definition]:
    """
    Parse a string containing one or more YAML model definitions.

    Args:
        source:  The file the content came from (to help with better logging)
        model_content:  The YAML to parse

    Returns:
        The AaC definitions that were built from the model contents.
    """

    def mark_to_source_location(start: Mark, end: Mark) -> SourceLocation:
        return SourceLocation(start.line, start.column, start.index, end.column - start.column)

    def get_lexemes_for_definition(value_tokens, content_start, content_end) -> list[Lexeme]:
        definition_tokens = [token for token in value_tokens if is_token_between_locations(token, content_start, content_end)]
        definition_lexemes = []
        for token in definition_tokens:
            location = mark_to_source_location(token.start_mark, token.end_mark)
            definition_lexemes.append(Lexeme(location, source, token.value))
        return definition_lexemes

    def is_token_between_locations(token, inclusive_line_start: int, inclusive_line_end: int) -> list[Lexeme]:
        return token.start_mark.line >= inclusive_line_start and token.end_mark.line <= inclusive_line_end

    yaml_tokens: list[Token] = YAML_CACHE.scan_string(model_content, source)
    value_tokens: list[Token] = [token for token in yaml_tokens if hasattr(token, "value")]
    doc_start_token: list[StreamStartToken] = [token for token in yaml_tokens if isinstance(token, StreamStartToken)]
    doc_end_token: list[StreamEndToken] = [token for token in yaml_tokens if isinstance(token, StreamEndToken)]
    doc_segment_tokens: list[DocumentStartToken] = [token for token in yaml_tokens if isinstance(token, DocumentStartToken)]
    doc_tokens = [*doc_start_token, *doc_segment_tokens, *doc_end_token]

    yaml_dicts: list[dict] = deepcopy(YAML_CACHE.parse_string(model_content, source))

    source_files: dict[str, AaCFile] = {}
    definitions: list[Definition] = []
    for doc_token_index in range(0, len(doc_tokens) - 1):
        start_doc_token = doc_tokens[doc_token_index]
        end_doc_token = doc_tokens[doc_token_index + 1]

        content_start_line = start_doc_token.start_mark.line
        content_end_line = end_doc_token.end_mark.line + (1 if isinstance(end_doc_token, StreamEndToken) else 0)

        yaml_text = linesep.join(model_content.splitlines()[content_start_line:content_end_line])
        yaml_text += linesep

        if yaml_text.strip():
            definition_lexemes = get_lexemes_for_definition(value_tokens, content_start_line, content_end_line)

            if yaml_dicts:
                source_file = source_files.get(source)
                if not source_file:
                    source_file = AaCFile(source, True, False)
                    source_files[source] = source_file

                root_yaml = yaml_dicts.pop(0)
                root_type, *_ = root_yaml.keys()
                definition_name = root_yaml.get(root_type, {}).get(DEFINITION_FIELD_NAME, "")
                definition_package = root_yaml.get(root_type, {}).get("package", "")

                new_definition = Definition(
                    name=definition_name,
                    package=definition_package,
                    content=yaml_text,
                    source=source_file,
                    lexemes=definition_lexemes,
                    structure=root_yaml,
                )
                definitions.append(new_definition)
        else:
            logging.info(
                f"Skipping empty content between {start_doc_token}:L{content_start_line} and {end_doc_token}:L{content_end_line} in source {source}"
            )
            logging.debug(f"Source: {source} Content:{model_content}")
            logging.debug(f"Content lines:{model_content.splitlines()}")

    return definitions


def _read_arch_file_content(arch_file: str) -> str:
    """
    Read file content method extracts text content from the specified architecture as code file.

    Args:
        arch_file: The file to read.

    Returns:
        The contents of the file as a string.
    """
    content = ""
    acceptable_file_extensions = [YAML_DOCUMENT_EXTENSION, AAC_DOCUMENT_EXTENSION]

    _, file_ext = path.splitext(arch_file)
    if file_ext in acceptable_file_extensions:
        try:
            with open(arch_file, "r") as file:
                content = file.read()
        except IOError as error:
            logging.error(f"Failed to parse {arch_file} with error {error}")

        if not content:
            logging.error(f"Failed to parse {arch_file}, it's an empty file.")
    else:
        logging.error(
            f"Failed to parse {arch_file}, the file is not an accepted Architecture-as-Code file extension {acceptable_file_extensions}."
        )

    return content


def get_definitions_by_root_key(root_key: str, definitions: list[Definition]) -> list[Definition]:
    """Return a subset of definitions with the given root key.

    The aac.in_out.parser.parse() function returns a list of all parsed Definitions.  Sometimes it is
    useful to only work with certain roots (i.e. model or schema).  This utility
    method allows the list of parsed definitions to be "filtered" to a specific root key.

    Args:
        root_key (str): The root key to filter on.
        definitions (list[Definition]): The list of parsed definitions to filter.

    Returns:
        A list of parsed AaC model Definitions with the given root key.
    """

    def does_definition_root_match(definition: Definition) -> bool:
        return root_key == definition.get_root_key()

    return [definition for definition in definitions if does_definition_root_match(definition)]


def _get_files_to_import_from_definitions(source_file_path: str, definitions: list[Definition]) -> list[str]:
    """
    Return a list of files to import from the list of definitions.

    This function assumes the list of definitions contains some import definitions, and it returns
    the list of file paths to import adjusted with the location of the source in case of relative import paths.
    """
    import_paths = set()

    import_definitions = get_definitions_by_root_key(ROOT_KEY_IMPORT, definitions)
    for definition in import_definitions:
        for import_path in definition.structure[ROOT_KEY_IMPORT]["files"] or []:
            arch_file_dir = path.dirname(path.realpath(source_file_path))
            parse_path = path.join(arch_file_dir, import_path.removeprefix(f".{path.sep}"))
            import_paths.add(sanitize_filesystem_path(parse_path))

    return list(import_paths)



"""This module manages a singleton instance of the YAML Parser Cache."""
from aac.in_out.parser._cache import YamlLFUCache


YAML_CACHE: YamlLFUCache = None


def get_cache() -> YamlLFUCache:
    """
    Return the YAML Parser Cache instance.

    Returns:
        The YAML Parser Cache instance
    """
    global YAML_CACHE

    if not YAML_CACHE:
        YAML_CACHE = YamlLFUCache()

    return YAML_CACHE


def reset_cache() -> None:
    """Resets the cache clearing all data."""
    global YAML_CACHE
    YAML_CACHE = YamlLFUCache()



from unittest import TestCase
from aac.execute.aac_execution_result import ExecutionResult, ExecutionMessage, ExecutionStatus, MessageLevel


class TestExecutionResult(TestCase):

    def test_execution_result_name(self):
        execution_result = ExecutionResult("name", "command_name", ExecutionStatus.SUCCESS, [])
        self.assertEqual(execution_result.plugin_name, "name")
        self.assertEqual(execution_result.plugin_command_name, "command_name")

    def test_execution_result_messages(self):
        execution_result = ExecutionResult("name", "command_name", ExecutionStatus.SUCCESS, [])
        execution_result.add_messages([
            ExecutionMessage("message 1", MessageLevel.INFO, None, None),
            ExecutionMessage("message 2", MessageLevel.INFO, None, None),
            ExecutionMessage("message 3", MessageLevel.INFO, None, None),
            ExecutionMessage("message 4", MessageLevel.INFO, None, None)
        ])

        messages = execution_result.get_messages_as_string()
        self.assertEqual(messages, "message 1\nmessage 2\nmessage 3\nmessage 4\n")

    def test_execution_result_message_levels(self):
        execution_result = ExecutionResult("name", "command_name", ExecutionStatus.SUCCESS, [])
        execution_result.add_messages([
            ExecutionMessage("message 1", MessageLevel.DEBUG, None, None),
            ExecutionMessage("message 2", MessageLevel.INFO, None, None),
            ExecutionMessage("message 3", MessageLevel.WARNING, None, None),
            ExecutionMessage("message 4", MessageLevel.ERROR, None, None)
        ])
        messages_list = execution_result.messages
        self.assertEqual(messages_list[0].level, MessageLevel.DEBUG)
        self.assertEqual(messages_list[1].level, MessageLevel.INFO)
        self.assertEqual(messages_list[2].level, MessageLevel.WARNING)
        self.assertEqual(messages_list[3].level, MessageLevel.ERROR)

    def test_execution_result_status_code(self):
        execution_result = ExecutionResult("name", "command_name", ExecutionStatus.GENERAL_FAILURE, [])
        self.assertFalse(execution_result.is_success())
        execution_result.status_code = ExecutionStatus.SUCCESS
        self.assertTrue(execution_result.is_success())

    def test_execution_result_status_enum(self):
        execution_success = ExecutionResult("name", "command_name", ExecutionStatus.SUCCESS, [])
        execution_constraint_failure = ExecutionResult("name", "command_name", ExecutionStatus.CONSTRAINT_FAILURE, [])
        execution_constraint_warning = ExecutionResult("name", "command_name", ExecutionStatus.CONSTRAINT_WARNING, [])
        execution_parser_failure = ExecutionResult("name", "command_name", ExecutionStatus.PARSER_FAILURE, [])
        execution_plugin_failure = ExecutionResult("name", "command_name", ExecutionStatus.PLUGIN_FAILURE, [])
        execution_operation_cancelled = ExecutionResult("name", "command_name", ExecutionStatus.OPERATION_CANCELLED, [])
        execution_general_failure = ExecutionResult("name", "command_name", ExecutionStatus.GENERAL_FAILURE, [])

        self.assertEqual(execution_success.status_code, ExecutionStatus.SUCCESS)
        self.assertEqual(execution_constraint_failure.status_code, ExecutionStatus.CONSTRAINT_FAILURE)
        self.assertEqual(execution_constraint_warning.status_code, ExecutionStatus.CONSTRAINT_WARNING)
        self.assertEqual(execution_parser_failure.status_code, ExecutionStatus.PARSER_FAILURE)
        self.assertEqual(execution_plugin_failure.status_code, ExecutionStatus.PLUGIN_FAILURE)
        self.assertEqual(execution_operation_cancelled.status_code, ExecutionStatus.OPERATION_CANCELLED)
        self.assertEqual(execution_general_failure.status_code, ExecutionStatus.GENERAL_FAILURE)

    def test_execution_result_message_levels_invalid(self):
        with self.assertRaises(TypeError):
            execution_result = ExecutionResult("name", "command_name", ExecutionStatus.SUCCESS, [])
            execution_result.add_message("message", "INFO", None, None)
        with self.assertRaises(AttributeError):
            execution_result = ExecutionResult("name", "command_name", ExecutionStatus.SUCCESS, [])
            execution_result.add_message("message", MessageLevel.info, None, None)

    def test_execution_result_invalid_name(self):
        with self.assertRaises(TypeError):
            execution_result = ExecutionResult(123, "command_name", ExecutionStatus.GENERAL_FAILURE, [])  # noqa: F841
        with self.assertRaises(TypeError):
            execution_result = ExecutionResult("name", 456, ExecutionStatus.GENERAL_FAILURE, [])  # noqa: F841

    def test_execution_result_invalid_status(self):
        with self.assertRaises(AttributeError):
            execution_result = ExecutionResult("name", "command_name", ExecutionStatus.general_failure, [])  # noqa: F841
        with self.assertRaises(TypeError):
            execution_result = ExecutionResult("name", "command_name", "general_failure", [])  # noqa: F841

    def test_execution_result_no_message(self):
        with self.assertRaises(TypeError):
            execution_result = ExecutionResult("name", "command_name", ExecutionStatus.SUCCESS, [])
            execution_result.add_message("message")



"""__init__.py module for the Constraint Assignment Arguments plugin."""
# WARNING - DO NOT EDIT - YOUR CHANGES WILL NOT BE PROTECTED.
# This file is auto-generated by the aac gen-plugin and may be overwritten.

from os.path import join, dirname
from typing import Any
from aac.execute.aac_execution_result import (
    ExecutionResult,
)
from aac.execute import hookimpl
from aac.context.language_context import LanguageContext
from aac.context.definition import Definition
from aac.execute.plugin_runner import PluginRunner


from aac.plugins.constraint_assignment_arguments.constraint_assignment_arguments_impl import (
    plugin_name,
)


from aac.plugins.constraint_assignment_arguments.constraint_assignment_arguments_impl import (
    check_arguments_against_constraint_definition,
)


constraint_assignment_arguments_aac_file_name = "constraint_assignment_arguments.aac"


def run_check_arguments_against_constraint_definition(
    instance: Any, definition: Definition, defining_schema, arguments: Any
) -> ExecutionResult:
    """Ensures the arguments provided in the assignment match the arguments defined in the constraint definition."""

    return check_arguments_against_constraint_definition(
        instance, definition, defining_schema
    )


@hookimpl
def register_plugin() -> None:
    """
    Returns information about the plugin.

    Returns:
        A collection of information about the plugin and what it contributes.
    """

    active_context = LanguageContext()
    constraint_assignment_arguments_aac_file = join(
        dirname(__file__), constraint_assignment_arguments_aac_file_name
    )
    definitions = active_context.parse_and_load(
        constraint_assignment_arguments_aac_file
    )

    constraint_assignment_arguments_plugin_definition = [
        definition for definition in definitions if definition.name == plugin_name
    ][0]

    plugin_instance = constraint_assignment_arguments_plugin_definition.instance
    for file_to_load in plugin_instance.definition_sources:
        active_context.parse_and_load(file_to_load)

    plugin_runner = PluginRunner(
        plugin_definition=constraint_assignment_arguments_plugin_definition
    )

    plugin_runner.add_constraint_callback(
        "Check arguments against constraint definition",
        run_check_arguments_against_constraint_definition,
    )

    active_context.register_plugin_runner(plugin_runner)



from os.path import basename, sep, isfile
from os import linesep
from tempfile import TemporaryDirectory
from typing import Optional
from unittest import TestCase
import tempfile

from aac.in_out.constants import AAC_DOCUMENT_EXTENSION, YAML_DOCUMENT_SEPARATOR
from aac.in_out.parser import parse, ParserError
from aac.context.language_context import LanguageContext
from aac.context.definition import Definition
from aac.context.source_location import SourceLocation

PARSER_TEST_SOURCE = "<parser test>"


class TestParser(TestCase):

    def create_test_file(self, path: str, content: str) -> str:
        with open(path, "w") as test_file:
            test_file.write(content)
        return path
    
    def get_lexeme_values_for_definition(self, name: str, definitions: list[Definition]) -> list[str]:
        lexemes = [definition.lexemes for definition in definitions if definition.name == name][0]
        return [lexeme.value for lexeme in lexemes]

    def check_definition(self, definition: Optional[Definition], name: str, type: str):
        self.assertIsNotNone(definition)
        self.assertEqual(type, definition.get_root_key())
        self.assertEqual(name, definition.name)

    def check_parser_errors(self, filespec: str, *error_messages: list[str]):
        with self.assertRaises(ParserError) as cm:
            parse(filespec)

        # Assert that each error message is contained in the returned error message string.
        errors = sep.join(cm.exception.errors)
        list(map(lambda e: self.assertIn(e, errors), error_messages))
        self.assertEqual(cm.exception.source, filespec)

    def test_can_get_lexemes_from_parsed_architecture_file(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            test_yaml = self.create_test_file(f"{temp_dir}/test.aac", TEST_MESSAGE_CONTENTS.strip())
            parsed_definitions = parse(test_yaml)
            self.assertEqual(1, len(parsed_definitions))
            parsed_definition = parsed_definitions[0]

            self.assertGreaterEqual(len(parsed_definition.lexemes), 2)

            first, second, *_ = parsed_definition.lexemes
            self.assertEqual(first.source, test_yaml)
            self.assertEqual(first.value, "schema")
            self.assertEqual(first.location, SourceLocation(0, 0, 0, 6))

            self.assertEqual(second.source, test_yaml)
            self.assertEqual(second.value, "name")
            self.assertEqual(second.location, SourceLocation(1, 2, 10, 4))

    def test_errors_when_parsing_invalid_yaml(self):
        content = "model: ]["
        with tempfile.TemporaryDirectory() as temp_dir:
            test_yaml = self.create_test_file(f"{temp_dir}/test.aac", content)
            self.check_parser_errors(test_yaml, "invalid YAML")

    def test_errors_when_parsing_incomplete_model(self):
        content = "model:"
        with tempfile.TemporaryDirectory() as temp_dir:
            test_yaml = self.create_test_file(f"{temp_dir}/test.aac", content)
            self.check_parser_errors(test_yaml, "Definition is missing field 'name'", content)

    def test_errors_when_parsing_non_yaml(self):
        content = "not yaml"
        with tempfile.TemporaryDirectory() as temp_dir:
            test_yaml = self.create_test_file(f"{temp_dir}/test.aac", content)
            self.check_parser_errors(test_yaml, "not YAML", content)

    def test_content_is_split_by_yaml_documents(self):
        content = f"{TEST_MESSAGE_CONTENTS}{YAML_DOCUMENT_SEPARATOR}{TEST_STATUS_CONTENTS}"
        parsed_definitions = parse(content, source_uri=PARSER_TEST_SOURCE)

        self.assertEqual(len(parsed_definitions), 2)

        contents = [definition.content for definition in parsed_definitions]
        self.assertIn(TEST_MESSAGE_CONTENTS, contents)
        self.assertIn(f"{YAML_DOCUMENT_SEPARATOR}{TEST_STATUS_CONTENTS}", contents)

    def test_file_content_is_split_by_yaml_documents(self):
        content = f"{TEST_MESSAGE_CONTENTS}{YAML_DOCUMENT_SEPARATOR}{TEST_STATUS_CONTENTS}"
        with tempfile.TemporaryDirectory() as temp_dir:
            test_yaml = self.create_test_file(f"{temp_dir}/test.aac", content)
            parsed_definitions = parse(test_yaml)

            self.assertEqual(len(parsed_definitions), 2)

            contents = [definition.content for definition in parsed_definitions]
            self.assertIn(TEST_MESSAGE_CONTENTS, contents)
            self.assertIn(f"{YAML_DOCUMENT_SEPARATOR}{TEST_STATUS_CONTENTS}", contents)

    def test_lexemes_are_split_by_yaml_documents(self):
        content = f"{TEST_MESSAGE_CONTENTS}{YAML_DOCUMENT_SEPARATOR}{TEST_STATUS_CONTENTS}"
        parsed_definitions = parse(content, source_uri=PARSER_TEST_SOURCE)

        self.assertEqual(len(parsed_definitions), 2)

        self.assertNotIn(
            TEST_STATUS_NAME,
            self.get_lexeme_values_for_definition(TEST_MESSAGE_NAME, parsed_definitions),
        )
        self.assertNotIn(
            TEST_MESSAGE_NAME,
            self.get_lexeme_values_for_definition(TEST_STATUS_NAME, parsed_definitions),
        )

    def test_file_lexemes_are_split_by_yaml_documents(self):
        content = f"{TEST_MESSAGE_CONTENTS}{YAML_DOCUMENT_SEPARATOR}{TEST_STATUS_CONTENTS}"
        with tempfile.TemporaryDirectory() as temp_dir:
            test_yaml = self.create_test_file(f"{temp_dir}/test.aac", content)
            parsed_definitions = parse(test_yaml)

            self.assertEqual(len(parsed_definitions), 2)

            self.assertNotIn(
                TEST_STATUS_NAME,
                self.get_lexeme_values_for_definition(TEST_MESSAGE_NAME, parsed_definitions),
            )
            self.assertNotIn(
                TEST_MESSAGE_NAME,
                self.get_lexeme_values_for_definition(TEST_STATUS_NAME, parsed_definitions),
            )

    def test_does_not_add_doc_separator_if_not_already_present(self):
        definition, *_ = parse(TEST_MESSAGE_CONTENTS, source_uri=PARSER_TEST_SOURCE)
        self.assertEqual(definition.content, TEST_MESSAGE_CONTENTS)

        definitions = parse(
            f"{TEST_MESSAGE_CONTENTS}{YAML_DOCUMENT_SEPARATOR}{TEST_STATUS_CONTENTS}", source_uri=PARSER_TEST_SOURCE
        )

        message_definition, *_ = [definition for definition in definitions if definition.name == TEST_MESSAGE_NAME]
        self.assertFalse(message_definition.content.startswith(YAML_DOCUMENT_SEPARATOR))

        status_definition, *_ = [definition for definition in definitions if definition.name == TEST_STATUS_NAME]
        self.assertTrue(status_definition.content.startswith(YAML_DOCUMENT_SEPARATOR))

    def test_no_errors_for_empty_definitions(self):
        definitions = parse(
            f"{TEST_MESSAGE_CONTENTS}{YAML_DOCUMENT_SEPARATOR}{linesep}{YAML_DOCUMENT_SEPARATOR}{TEST_STATUS_CONTENTS}"
        )
        self.assertEqual(len(definitions), 2)


class TestParserImports(TestCase):
    """Exercises the import functionality of the parser."""

    def create_test_file(self, path: str, content: str) -> str:
        with open(path, "w") as test_file:
            test_file.write(content)
        return path

    def check_definition(self, definition: Optional[Definition], name: str, type: str):
        self.assertIsNotNone(definition)
        self.assertEqual(type, definition.get_root_key())
        self.assertEqual(name, definition.name)

    def test_can_handle_architecture_file_with_imports(self):
        with TemporaryDirectory() as temp_dir:
            import1 = self.create_test_file(f"{temp_dir}/{TEST_MESSAGE_FILE_NAME}", TEST_MESSAGE_CONTENTS)
            import2 = self.create_test_file(f"{temp_dir}/{TEST_STATUS_FILE_NAME}", TEST_STATUS_CONTENTS)
            test_yaml = self.create_test_file(f"{temp_dir}/test_model.aac", TEST_MODEL_CONTENTS)

            definitions = parse(test_yaml)
            self.assertEqual(len(definitions), 4)

            context = LanguageContext()
            definitions = context.parse_and_load(test_yaml)
            self.assertEqual(len(definitions), 4)

            import_definitions = [definition for definition in definitions if definition.get_root_key() == "import"]
            import_basenames = []
            for imp in import_definitions:
                import_basenames.extend([basename(entry) for entry in imp.instance.files])
            self.assertIn(basename(import1), import_basenames)
            self.assertIn(basename(import2), import_basenames)

            schema_message_definition = context.get_definitions_by_name(TEST_MESSAGE_NAME)[0]
            enum_status_definition = context.get_definitions_by_name(TEST_STATUS_NAME)[0]
            model_echosvc_definition = context.get_definitions_by_name(TEST_MODEL_NAME)[0]

            self.check_definition(schema_message_definition, TEST_MESSAGE_NAME, "schema")
            self.check_definition(enum_status_definition, TEST_STATUS_NAME, "enum")
            self.check_definition(model_echosvc_definition, TEST_MODEL_NAME, "model")

            self.assertEqual(schema_message_definition.source.uri, import1)
            self.assertEqual(enum_status_definition.source.uri, import2)
            self.assertEqual(model_echosvc_definition.source.uri, test_yaml)

            first_import, *_ = [definition for definition in definitions if definition.get_root_key() == "import"]
            first, _, third, *_ = first_import.lexemes
            self.assertEqual(first.source, test_yaml)
            self.assertEqual(first.value, "import")
            self.assertEqual(first.location, SourceLocation(1, 0, 1, 6))

            self.assertEqual(third.source, test_yaml)
            self.assertEqual(third.value, f"./{TEST_MESSAGE_FILE_NAME}")
            self.assertEqual(third.location, SourceLocation(3, 6, 24, 2 + len(TEST_MESSAGE_FILE_NAME)))

    def test_handles_multiple_import_sections_per_file(self):
        another_definition = """
enum:
    name: TestEnum
    values:
    - a
    - b
    - c
"""
        with TemporaryDirectory() as temp_dir:
            import1 = self.create_test_file(f"{temp_dir}/{TEST_STATUS_FILE_NAME}", another_definition)
            import2 = self.create_test_file(f"{temp_dir}/{TEST_MESSAGE_FILE_NAME}", TEST_MESSAGE_CONTENTS)
            import3 = self.create_test_file(f"{temp_dir}/{TEST_STATUS_FILE_NAME}", TEST_STATUS_CONTENTS)
            
            another_import_definition = f"""
import:
  files:
    - {import1}
"""
            content = f"{another_import_definition}{YAML_DOCUMENT_SEPARATOR}{TEST_MODEL_CONTENTS}"
            main = self.create_test_file(f"{temp_dir}/echo_service.aac", content)
            context = LanguageContext()
            definitions = context.parse_and_load(main)
            self.assertEqual(len(definitions), 5)

            import_def_1, import_def_2, *_ = [definition for definition in definitions if definition.get_root_key() == "import"]
            import_basenames = [basename(imp) for imp in import_def_1.instance.files + import_def_2.instance.files]
            self.assertIn(basename(import1), import_basenames)
            self.assertIn(basename(import2), import_basenames)
            self.assertIn(basename(import3), import_basenames)

    def test_parsed_file_imports_non_existent_file(self):
        non_existent_import_definition = """
import:
  files:
    - non_existent_file.aac
"""
        test_message_contents_with_bad_import = YAML_DOCUMENT_SEPARATOR.join([non_existent_import_definition, TEST_MESSAGE_CONTENTS])
        with TemporaryDirectory() as temp_dir:
            initial_file = self.create_test_file(f"{temp_dir}/{TEST_MESSAGE_FILE_NAME}", test_message_contents_with_bad_import)
            definitions = parse(initial_file)
            import_definitions = [definition for definition in definitions if definition.get_root_key() == "import"]

            """ Asserting that the bad import definition was parsed and that this assertion was allowed to complete without error is the
                assertion that the parse mechanism softly handled the bad import. """
            self.assertEqual(len(import_definitions), 1)

    def test_parsed_file_imports_empty_file(self):
        empty_file_name = "empty_file.aac"
        empty_import_content = """
import:
  files:

"""
        # non_existent_import_definition = create_import_definition([empty_file_name])
        test_message_contents_with_bad_import = YAML_DOCUMENT_SEPARATOR.join([empty_import_content, TEST_MESSAGE_CONTENTS])
        with TemporaryDirectory() as temp_dir:
            initial_file = self.create_test_file(f"{temp_dir}/{TEST_MESSAGE_FILE_NAME}", test_message_contents_with_bad_import)
            empty_file = self.create_test_file(f"{temp_dir}/{empty_file_name}", "")
            
            definitions = parse(initial_file)
            import_definitions = [definition for definition in definitions if definition.get_root_key() == "import"]

            self.assertTrue(isfile(empty_file))

            """ Asserting that the bad import definition was parsed and that this assertion was allowed to complete without error is the
                assertion that the parse mechanism softly handled the bad import. """
            self.assertEqual(len(import_definitions), 1)

    def test_parsed_file_imports_invalid_file(self):
        invalid_yaml_file_name = "invalid_yaml_file.aac"
        imvalid_yaml_file_content = f"""
import:
  files:
    - {invalid_yaml_file_name}
"""
        test_message_contents_with_bad_import = YAML_DOCUMENT_SEPARATOR.join([imvalid_yaml_file_content, TEST_MESSAGE_CONTENTS])
        with TemporaryDirectory() as temp_dir:
            initial_file = self.create_test_file(f"{temp_dir}/{TEST_MESSAGE_FILE_NAME}", test_message_contents_with_bad_import)
            empty_file = self.create_test_file(f"{temp_dir}/{invalid_yaml_file_name}", "")
            definitions = parse(initial_file)
            import_definitions = [definition for definition in definitions if definition.get_root_key() == "import"]

            self.assertTrue(isfile(empty_file))

            """ Asserting that the bad import definition was parsed and that this assertion was allowed to complete without error is the
                assertion that the parse mechanism softly handled the bad import. """
            self.assertEqual(len(import_definitions), 1)


TEST_MODEL_NAME = "EchoService"
TEST_MESSAGE_NAME = "Message"
TEST_STATUS_NAME = "Status"

TEST_MESSAGE_FILE_NAME = f"{TEST_MESSAGE_NAME}{AAC_DOCUMENT_EXTENSION}"
TEST_STATUS_FILE_NAME = f"{TEST_STATUS_NAME}{AAC_DOCUMENT_EXTENSION}"

TEST_MESSAGE_CONTENTS = f"""
schema:
  name: {TEST_MESSAGE_NAME}
  package: test_aac.in_out
  fields:
  - name: body
    type: string
  - name: sender
    type: string
"""
TEST_STATUS_CONTENTS = f"""
enum:
  name: {TEST_STATUS_NAME}
  package: test_aac.in_out
  values:
    - sent
    - 'failed to send'
"""
TEST_MODEL_CONTENTS = f"""
import:
  files:
    - .{sep}{TEST_MESSAGE_FILE_NAME}
    - {TEST_STATUS_FILE_NAME}
---
model:
  name: {TEST_MODEL_NAME}
  description: This is a message mirror.
  behavior:
    - name: echo
      description: This is the one thing it does.
      input:
        - name: inbound
          type: Message
      output:
        - name: outbound
          type: Message
      acceptance:
        - name: this is a test
          scenarios: 
          - name: onReceive
            given:
            - The EchoService is running.
            when:
                - The user sends a message to EchoService.
            then:
                - The user receives the same message from EchoService.
"""



"""Provides lexical unit for a parsed AaC definition."""

from attr import attrib, attrs, validators

from aac.in_out.paths import is_same_file
from aac.context.source_location import SourceLocation


@attrs(eq=False)
class Lexeme:
    """A lexical unit for a parsed AaC definition.

    Attributes:
        location (SourceLocation): The location at which the object was found.
        source (str): The source in which the object was found.
        value (str): The value of the parsed object.
    """

    location: SourceLocation = attrib(validator=validators.instance_of(SourceLocation))
    source: str = attrib(validator=validators.instance_of(str))
    value: str = attrib(validator=validators.instance_of(str))

    def __init__(self, location: SourceLocation, source: str, value: str):
        """
        An init constructor for the Lexeme Class.

        Args:
            location (SourceLocation): The position in the source file.
            source (str): The source file.
            value (str): The value stored in the lexeme.
        """
        self.location = location
        self.source = source
        self.value = value

    def __eq__(self, __o) -> bool:
        """Return whether this Lexeme is the same as __o."""
        return (
            isinstance(__o, Lexeme)
            and self.value == __o.value
            and self.location == __o.location
            and is_same_file(self.source, __o.source)
        )

    def __str__(self) -> str:
        """Return a string representation of this Lexeme."""
        return f"{self.value} at {self.location} in {self.source}"



from unittest import TestCase

from aac.context.language_context import LanguageContext
from aac.context.language_error import LanguageError

class TestLanguageError(TestCase):
    def test_language_error(self):
        try:
            raise LanguageError("message", "location")
        except LanguageError as e:
            self.assertEqual(str(e), "Error: message. Location: location")

        try:
            raise LanguageError(message="message", location="location")
        except LanguageError as e:
            self.assertEqual(str(e), "Error: message. Location: location")



from unittest import TestCase
from aac.context.language_context import LanguageContext
from aac.plugins.no_ext_for_final.no_ext_for_final_impl import no_extension_for_final


class TestNoExtForFinal(TestCase):

    #  Nominal (expected pass) test with a single valid Schema definition
    def test_no_extension_for_final_nominal(self):
        context = LanguageContext()
        schema_definition = context.get_definitions_by_name("aac.lang.Schema")
        if len(schema_definition) != 1:
            self.fail("Expected to find one and only one Schema definition")
        schema_definition = schema_definition[0]

        definitions = context.parse_and_load(root_schema_with_good_ext)
        result = no_extension_for_final(definitions[0].instance, definitions[0], schema_definition.instance)
        self.assertTrue(result.is_success())
        result = no_extension_for_final(definitions[1].instance, definitions[1], schema_definition.instance)
        self.assertTrue(result.is_success())
        context.remove_definitions(definitions)

    #  Off-nominal (expected fail) tests
    def test_no_extension_for_final_offnominal(self):
        context = LanguageContext()
        schema_definition = context.get_definitions_by_name("aac.lang.Schema")
        if len(schema_definition) != 1:
            self.fail("Expected to find one and only one Schema definition")
        schema_definition = schema_definition[0]

        #  Off-nominal (expected fail) test with a circular schema definition
        #  TestSchema tries to extend TestSchema
        definitions = context.parse_and_load(root_schema_with_bad_ext_circular)
        result = no_extension_for_final(definitions[0].instance, definitions[0], schema_definition.instance)
        self.assertTrue(result.is_success())
        result = no_extension_for_final(definitions[1].instance, definitions[1], schema_definition.instance)
        self.assertFalse(result.is_success())
        self.assertEqual(result.messages[0].source, "No file to reference")
        context.remove_definitions(definitions)

        #  Off-nominal (expected fail) test with a invalid schema definition
        #  TestSchema defined twice
        definitions = context.parse_and_load(root_schema_with_bad_ext_duplicate)
        result = no_extension_for_final(definitions[0].instance, definitions[0], schema_definition.instance)
        self.assertTrue(result.is_success())
        result = no_extension_for_final(definitions[1].instance, definitions[1], schema_definition.instance)
        self.assertFalse(result.is_success())

        #  Off-nominal (expected fail) test with a invalid schema definition
        #  Specified parent to extend does not exist (TestDoesNotExist)
        definitions = context.parse_and_load(root_schema_with_bad_ext_does_not_exist)
        result = no_extension_for_final(definitions[0].instance, definitions[0], schema_definition.instance)
        self.assertTrue(result.is_success())
        result = no_extension_for_final(definitions[1].instance, definitions[1], schema_definition.instance)
        self.assertFalse(result.is_success())
        context.remove_definitions(definitions)


root_schema_with_good_ext = """
schema:
  name: TestParent
  package: test_aac.plugins.no_ext_for_final
  fields:
    - name: name
      type: string
    - name: test_field
      type: string
---
schema:
  name: TestChild
  package: test_aac.plugins.no_ext_for_final
  extends:
    - name: TestParent
      package: test_aac.plugins.no_ext_for_final
  fields:
    - name: name
      type: string
    - name: test_field
      type: string
"""

root_schema_with_bad_ext_circular = """
schema:
  name: TestParent
  package: test_aac.plugins.no_ext_for_final
  modifiers:
    - final
  fields:
    - name: name
      type: string
    - name: test_field
      type: string
---
schema:
  name: TestChild
  package: test_aac.plugins.no_ext_for_final
  extends:
    - name: TestParent
      package: test_aac.plugins.no_ext_for_final
  fields:
    - name: name
      type: string
    - name: test_field
      type: string
"""

root_schema_with_bad_ext_duplicate = """
schema:
  name: TestParent
  package: test_aac.plugins.no_ext_for_final
  modifiers:
    - final
  fields:
    - name: name
      type: string
    - name: test_field
      type: string
---
schema:
  name: TestParent
  package: test_aac.plugins.no_ext_for_final
  extends:
    - name: TestChild
      package: test_aac.plugins.no_ext_for_final
  fields:
    - name: name
      type: string
    - name: test_field
      type: string
"""

root_schema_with_bad_ext_does_not_exist = """
schema:
  name: TestParent
  package: test_aac.plugins.no_ext_for_final
  fields:
    - name: name
      type: string
    - name: test_field
      type: string
---
schema:
  name: TestChild
  package: test_aac.plugins.no_ext_for_final
  extends:
    - name: TestDoesNotExist
      package: test_aac.plugins.no_ext_for_final
  fields:
    - name: name
      type: string
    - name: test_field
      type: string
"""



from unittest import TestCase
from aac.context.language_context import LanguageContext
from aac.context.language_error import LanguageError
from aac.in_out.parser._parser_error import ParserError
import traceback


from aac.plugins.root_schema_must_have_name.root_schema_must_have_name_impl import (
    plugin_name,
)


from aac.plugins.root_schema_must_have_name.root_schema_must_have_name_impl import (
    root_schema_has_name,
)


class TestRootSchemaMustHaveName(TestCase):

    # Happy path test
    def test_root_schema_has_name(self):
        context = LanguageContext()
        definitions = context.parse_and_load(root_schema_with_name)
        result = root_schema_has_name(definitions[0].instance, definitions[0], context.get_definitions_by_name("Schema")[0].instance)
        self.assertTrue(result.is_success())
        context.remove_definitions(definitions)


    #  This test starts with a normal properly formed schema
    #  It then takes the result from parse_and_load and specifically removes the 'name'
    #  It then calls root_schema_has_name which will not find the name it is looking for
    #  and return an ExecutionStatus.GENERAL_FAILURE from root_schema_has_name and a specific error_msg (see line 61)
    def test_root_schema_has_name_fail(self):
        context = LanguageContext()

        # Obtain the definitions for a properly formed schema
        # This prevents any ParserErrors or LanguageErrors from being generated within parse_and_load
        definitions = context.parse_and_load(root_schema_with_name)

        # Now modify the fields to remove the 'name' field
        # Most of this code is copied from root_schema_has_name::_get_fields
        schema = definitions[0].instance
        instance_fields = list[str]
        context: LanguageContext = LanguageContext()
        instance_fields: list[str] = []

        for field in schema.fields:
            if field.name != 'name':
                # Copy the field over except
                # skip if it is 'name' to trigger the check in root_schema_has_name
                instance_fields.append(field)

        if schema.extends:
            for ext in schema.extends:
                parent_schema = context.get_definitions_by_name(ext.name)
                if len(parent_schema) == 1:
                    definitions[0].instance.fields.extend(_get_fields(parent_schema[0].instance))
        ## Set fields to the modified list without a 'name' field
        definitions[0].instance.fields = instance_fields

        result = root_schema_has_name(definitions[0].instance, definitions[0], context.get_definitions_by_name("Schema")[0].instance)
        index = result.get_messages_as_string().find("must have a field named 'name'") # this is the message from line 61 of root_schema_has_name
        if index != -1:
            # Name message located
            self.assertFalse(result.is_success())
        else:
            # Name message missing
            self.assertFalse(True,"Error handling has been modified. Test FAILED.")
        context.remove_definitions(definitions)


    #  This test starts with a normal properly formed schema
    #  It then adds a second schema to the context with a name-collision
    #  It then calls root_schema_has_name which will find a non-unique definition and trigger a LanguageError in root_schema_has_name::_get_fields
    def test_root_schema_has_name(self):
        with self.assertRaises(LanguageError):

            context = LanguageContext()
            definitions = context.parse_and_load(root_schema_with_extends)
            context.parse_and_load(schema_root_with_different_package)

            result = root_schema_has_name(definitions[0].instance, definitions[0], context.get_definitions_by_name("Schema")[0].instance)
            context.remove_definitions(definitions)

    #  This test should trigger a LanguageError exception from the parse_and_load call
    #  So the only pass here is throwing a LanguageError
    def test_root_schema_has_name_language_error1(self):
        with self.assertRaises(LanguageError):

            context = LanguageContext()
            definitions = context.parse_and_load(root_schema_without_name_language_error1)
            result = root_schema_has_name(definitions[0].instance, definitions[0], context.get_definitions_by_name("Schema")[0].instance)
            context.remove_definitions(definitions)


    #  This test should trigger a LanguageError exception from the parse_and_load call
    #  So the only pass here is throwing a LanguageError
    def test_root_schema_has_name_language_error2(self):
        with self.assertRaises(LanguageError):

            context = LanguageContext()
            definitions = context.parse_and_load(root_schema_without_name_language_error2)
            result = root_schema_has_name(definitions[0].instance, definitions[0], context.get_definitions_by_name("Schema")[0].instance)
            context.remove_definitions(definitions)


    #  This test should trigger a ParserError exception from the parse_and_load call
    #  So the only pass here is throwing a ParserError
    def test_root_schema_has_name_parser_error3(self):
        with self.assertRaises(ParserError):

            context = LanguageContext()
            definitions = context.parse_and_load(root_schema_without_name_parser_error3)
            result = root_schema_has_name(definitions[0].instance, definitions[0], context.get_definitions_by_name("Schema")[0].instance)
            context.remove_definitions(definitions)


    #  This test should trigger a ParserError exception from the parse_and_load call
    #  So the only pass here is the self.assertFalse(False) in the ParserError catch block
    def test_root_schema_has_name_parser_error4(self):
        with self.assertRaises(ParserError):

            context = LanguageContext()
            definitions = context.parse_and_load(root_schema_without_name_parser_error4)
            result = root_schema_has_name(definitions[0].instance, definitions[0], context.get_definitions_by_name("Schema")[0].instance)
            context.remove_definitions(definitions)

root_schema_with_name = """
schema:
  name: test_schema
  root: test_root
  fields:
    - name: name
      type: string
    - name: test_field
      type: string
"""

root_schema_without_name_language_error1 = """
schema:
  name: test_name
  root: test_root
  fields:
    - name:
      type: string
"""

root_schema_without_name_language_error2 = """
schema:
  name: test_name
  root: test_root
  fields:
      type: string
"""

root_schema_without_name_parser_error3 = """
schema:
  name:
  root: test_root
  fields:
    - name: name
      type: string
"""

root_schema_without_name_parser_error4 = """
schema:
  root: test_root
  fields:
    - name: name
      type: string
"""

root_schema_with_extends = """
schema:
  name: Schema
  extends:
    - package: aac.lang
      name: AacType
  package: aac.lang
  root: schema
  description: |
    A definition that defines the schema/structure of data.
  fields:
    - name: extends
      type: SchemaExtension[]
      description: |
        A list of Schema definition names that this definition inherits from.
    - name: test
      type: SchemaExtension[]
      description: |
        A list of Schema definition names that this definition inherits from.
    - name: modifiers
      type: dataref(modifier.name)[]
      description:
        A means of further defining the schema and how it can be used within the model.
    - name: root
      type: string
      description: |
        The root key to use when declaring an instance of the type in yaml/aac files.
    - name: fields
      type: Field[]
      description: |
        A list of fields that make up the definition.
      is_required: true
    - name: requirements
      type: dataref(req.id)[]
      description: |
        A list of requirements associated with this schema.
    - name: constraints
      type: SchemaConstraintAssignment[]
"""
schema_root_with_different_package = """
schema:
  name: AacType
  package: wrong_package
  description: |
    A definition that defines the schema/structure of data.
  fields:
    - name: name
      type: string
    - name: test_field
      type: string
"""



"""__init__.py module for the Version plugin."""
# WARNING - DO NOT EDIT - YOUR CHANGES WILL NOT BE PROTECTED.
# This file is auto-generated by the aac gen-plugin and may be overwritten.

from os.path import join, dirname
from aac.execute.aac_execution_result import (
    ExecutionResult,
    ExecutionStatus,
)
from aac.execute import hookimpl
from aac.context.language_context import LanguageContext
from aac.execute.plugin_runner import PluginRunner
from aac.plugins.version.version_impl import plugin_name, version

version_aac_file_name = "version.aac"


def run_version() -> ExecutionResult:
    """Print the AaC package version."""

    result = ExecutionResult(plugin_name, "version", ExecutionStatus.SUCCESS, [])

    version_result = version()
    if not version_result.is_success():
        return version_result
    else:
        result.add_messages(version_result.messages)

    return result


@hookimpl
def register_plugin() -> None:
    """
    Returns information about the plugin.

    Returns:
        A collection of information about the plugin and what it contributes.
    """

    active_context = LanguageContext()
    version_aac_file = join(dirname(__file__), version_aac_file_name)
    definitions = active_context.parse_and_load(version_aac_file)

    version_plugin_definition = [
        definition for definition in definitions if definition.name == plugin_name
    ][0]

    plugin_instance = version_plugin_definition.instance
    for file_to_load in plugin_instance.definition_sources:
        active_context.parse_and_load(file_to_load)

    plugin_runner = PluginRunner(plugin_definition=version_plugin_definition)
    plugin_runner.add_command_callback("version", run_version)

    active_context.register_plugin_runner(plugin_runner)



"""Defines the AaC plugin interface via Pluggy Hookspecs."""

from aac.execute import hookspec


@hookspec
def register_plugin():
    """
    Returns information about the plugin.

    Returns:
        A collection of information about the plugin and what it contributes.
    """



from unittest import TestCase
from aac.in_out.parser import get_cache
from aac.in_out.parser._cache import YamlLFUCache
from aac.context.language_context import LanguageContext

MODEL_TEMPLATE = """
model:
  name: __name__
  description: This is a test model with no content.
"""


class TestParserCache(TestCase):
    def test_get_cache_has_core_spec(self):
        parser_cache = get_cache()

        context = LanguageContext()
        expected_spec_definitions = [definition.structure for definition in context.get_aac_core_definitions()]
        actual_spec_definitions_from_file = parser_cache.parse_file(context.get_aac_core_file_path())
        actual_spec_definitions_from_string = parser_cache.parse_string(context.get_aac_core_as_yaml())

        # Assert parse via filename is equivalent to expected core spec
        self.assertListEqual(expected_spec_definitions, actual_spec_definitions_from_file)

        # Assert parse via file content is equivalent to expected core spec
        self.assertListEqual(expected_spec_definitions, actual_spec_definitions_from_string)

    def test_cache_parse_ejects_least_frequently_used_content(self):
        parser_capacity = 4
        parser_cache = YamlLFUCache(capacity=parser_capacity)

        used_definition_1_name = "used_definition_1"
        used_definition_1 = MODEL_TEMPLATE.replace("__name__", used_definition_1_name)

        used_definition_2_name = "used_definition_2"
        used_definition_2 = MODEL_TEMPLATE.replace("__name__", used_definition_2_name)

        used_definition_3_name = "used_definition_3"
        used_definition_3 = MODEL_TEMPLATE.replace("__name__", used_definition_3_name)

        used_definition_4_name = "used_definition_4"
        used_definition_4 = MODEL_TEMPLATE.replace("__name__", used_definition_4_name)

        unused_definition_1_name = "unused_definition_1"
        unused_definition_1 = MODEL_TEMPLATE.replace("__name__", unused_definition_1_name)

        # Assert the cache is empty at the beginning
        self.assertEqual(0, len(parser_cache.cache))

        # Parse and cache the test strings
        parsed_unused_definition_1 = parser_cache.parse_string(unused_definition_1)[0]
        parser_cache.parse_string(used_definition_1)
        parser_cache.parse_string(used_definition_2)
        parser_cache.parse_string(used_definition_3)

        # Assert the cache is full at the end
        self.assertEqual(parser_capacity, len(parser_cache.cache))

        # Re-parse the used definitions to bump their use count per parse
        parser_cache.parse_string(used_definition_1)
        parser_cache.parse_string(used_definition_1)
        parser_cache.parse_string(used_definition_1)

        parser_cache.parse_string(used_definition_2)
        parser_cache.parse_string(used_definition_2)

        parser_cache.parse_string(used_definition_3)

        # Grab and sort the parser cache entries since their stored internally in an unordered hash map.
        cached_values = parser_cache._get_entries_sorted_by_access_count()
        # Assert the order of the cached entries
        self.assertEqual(3, cached_values[0].times_accessed)
        self.assertEqual(2, cached_values[1].times_accessed)
        self.assertEqual(1, cached_values[2].times_accessed)
        self.assertEqual(0, cached_values[3].times_accessed)

        # Assert that the last element in the cache is the unused definition
        expected_unused_entry = parser_cache._get_entries_sorted_by_access_count().pop(-1)
        expected_unused_dict, *_ = expected_unused_entry.yaml_structures
        self.assertDictEqual(expected_unused_dict, parsed_unused_definition_1)

        # Parse the new used string
        parsed_used_definition_4 = parser_cache.parse_string(used_definition_4)[0]

        # Assert that the last element in the cache is the last used definition
        expected_used_entry = parser_cache._get_entries_sorted_by_access_count().pop(-1)
        expected_used_dict, *_ = expected_used_entry.yaml_structures
        self.assertDictEqual(expected_used_dict, parsed_used_definition_4)

        # Assert the cache is full at the end
        self.assertEqual(parser_capacity, len(parser_cache.cache))
